#!/usr/bin/env python3
#
# Almost completely copied from the Kodi BBC addon, which has an MIT license, retained for the modifications:
# https://github.com/jonjomckay/kodi-addon-bbcsounds
# Copyright 2020 Jonjo McKay
#
# Additional code and modifications:
# Copyright (C) 2021 J.F.Dockes
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
# associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
# NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT
# OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import datetime
import dateutil.parser
import json
import os
import re
import requests
import sys
import threading
from time import mktime
import urllib

try:
    import feedparser

    hasfeedparser = True
except:
    hasfeedparser = False
from bs4 import BeautifulSoup

import cmdtalkplugin
from upmplgutils import *
from upmplgmodels import *
from xbmcplug import *

# Reusing the Tidal plugin routing module
from routing import Plugin

plugin = Plugin("")

bbcidprefix = "0$bbc$"

# I get a list of episodes from the web page, some of which are unavailable, and I can't know at
# this point.
#
# I then retrieve more detailed data for each episode, which takes time. At this point I know if an
# episode is available or not.
#
# The problem is that I get a total count from the Web page and return it to the CP. The count ends
# up being false because of the unavailable episodes.  Bubble and Upplay just deal with it and
# display the entries they actually got. Kazoo insists on displaying the total count, displaying an
# "unavailable" messages for the missing entries.
#
# Also we get duplicate entries on further slices because we can't know how many to skip without
# asking for the details (which would void the point), so that we don't skip enough.
#
# So actually using offset and count does not really work, and does not gain so much time anyway
# (because of the plgwithslave cache: this would only be useful if the client actually did not read
# the whole list).
#
# We just do as for most of the other plugins, and just read the whole data on each
# requests. Plgwithslave caches it and returns the appropriate slice to the client.
# Kept the code around anyway, but it's disabled by the following switch.
_g_do_slices = False

# Cmdtalk communication with our parent process
# Cmdtalk func name to method mapper
dispatcher = cmdtalkplugin.Dispatch()
# Cmdtalk pipe message handler
msgproc = cmdtalkplugin.Processor(dispatcher)

stations = [
    {
        "id": "p00fzl68",
        "name": "BBC Asian Network",
        "image": "bbc_asian_network_colour",
    },
    {
        "id": "p00fzl78",
        "name": "BBC Coventry & Warwickshire",
        "image": "bbc_radio_coventry_warwickshire_colour",
    },
    {"id": "p00fzl7f", "name": "BBC Essex", "image": "bbc_radio_essex_colour"},
    {
        "id": "p00fzl7q",
        "name": "BBC Hereford & Worcester",
        "image": "bbc_radio_hereford_worcester_colour",
    },
    {"id": "p00fzl82", "name": "BBC Newcastle", "image": "bbc_radio_newcastle_colour"},
    {"id": "p00fzl86", "name": "BBC Radio 1", "image": "bbc_radio_one_colour"},
    {"id": "p00fzl64", "name": "BBC Radio 1Xtra", "image": "bbc_1xtra_colour"},
    {"id": "p00fzl8v", "name": "BBC Radio 2", "image": "bbc_radio_two_colour"},
    {"id": "p00fzl8t", "name": "BBC Radio 3", "image": "bbc_radio_three_colour"},
    {"id": "p00fzl7j", "name": "BBC Radio 4 FM", "image": "bbc_radio_fourfm_colour"},
    {"id": "p00fzl7k", "name": "BBC Radio 4 LW", "image": "bbc_radio_four_colour"},
    {
        "id": "p00fzl7l",
        "name": "BBC Radio 4 Extra",
        "image": "bbc_radio_four_extra_colour",
    },
    {
        "id": "p00fzl7g",
        "name": "BBC Radio 5 live",
        "image": "bbc_radio_five_live_colour",
    },
    {
        "id": "p00fzl7h",
        "name": "BBC Radio 5 live sports extra",
        "image": "bbc_radio_five_live_sports_extra_colour",
    },
    {"id": "p00fzl65", "name": "BBC Radio 6 Music", "image": "bbc_6music_colour"},
    {
        "id": "p00fzl74",
        "name": "BBC Radio Berkshire",
        "image": "bbc_radio_berkshire_colour",
    },
    {
        "id": "p00fzl75",
        "name": "BBC Radio Bristol",
        "image": "bbc_radio_bristol_colour",
    },
    {
        "id": "p00fzl76",
        "name": "BBC Radio Cambridgeshire",
        "image": "bbc_radio_cambridge_colour",
    },
    {
        "id": "p00fzl77",
        "name": "BBC Radio Cornwall",
        "image": "bbc_radio_cornwall_colour",
    },
    {
        "id": "p00fzl79",
        "name": "BBC Radio Cumbria",
        "image": "bbc_radio_cumbria_colour",
    },
    {"id": "p00fzl7b", "name": "BBC Radio Cymru", "image": "bbc_radio_cymru_colour"},
    {"id": "p00fzl7c", "name": "BBC Radio Derby", "image": "bbc_radio_derby_colour"},
    {"id": "p00fzl7d", "name": "BBC Radio Devon", "image": "bbc_radio_devon_colour"},
    {"id": "p00fzl7m", "name": "BBC Radio Foyle", "image": "bbc_radio_foyle_colour"},
    {
        "id": "p00fzl7n",
        "name": "BBC Radio Gloucestershire",
        "image": "bbc_radio_gloucestershire_colour",
    },
    {
        "id": "p00fzl7p",
        "name": "BBC Radio Guernsey",
        "image": "bbc_radio_guernsey_colour",
    },
    {
        "id": "p00fzl7r",
        "name": "BBC Radio Humberside",
        "image": "bbc_radio_humberside_colour",
    },
    {"id": "p00fzl7s", "name": "BBC Radio Jersey", "image": "bbc_radio_jersey_colour"},
    {"id": "p00fzl7t", "name": "BBC Radio Kent", "image": "bbc_radio_kent_colour"},
    {
        "id": "p00fzl7v",
        "name": "BBC Radio Lancashire",
        "image": "bbc_radio_lancashire_colour",
    },
    {"id": "p00fzl7w", "name": "BBC Radio Leeds", "image": "bbc_radio_leeds_colour"},
    {
        "id": "p00fzl7x",
        "name": "BBC Radio Leicester",
        "image": "bbc_radio_leicester_colour",
    },
    {
        "id": "p00fzl7y",
        "name": "BBC Radio Lincolnshire",
        "image": "bbc_radio_lincolnshire_colour",
    },
    {"id": "p00fzl6f", "name": "BBC Radio London", "image": "bbc_london_colour"},
    {
        "id": "p00fzl7z",
        "name": "BBC Radio Manchester",
        "image": "bbc_radio_manchester_colour",
    },
    {
        "id": "p00fzl80",
        "name": "BBC Radio Merseyside",
        "image": "bbc_radio_merseyside_colour",
    },
    {
        "id": "p00fzl81",
        "name": "BBC Radio Nan Gaidheal",
        "image": "bbc_radio_nan_gaidheal_colour",
    },
    {
        "id": "p00fzl83",
        "name": "BBC Radio Norfolk",
        "image": "bbc_radio_norfolk_colour",
    },
    {
        "id": "p00fzl84",
        "name": "BBC Radio Northampton",
        "image": "bbc_radio_northampton_colour",
    },
    {
        "id": "p00fzl85",
        "name": "BBC Radio Nottingham",
        "image": "bbc_radio_nottingham_colour",
    },
    {"id": "p00fzl8c", "name": "BBC Radio Oxford", "image": "bbc_radio_oxford_colour"},
    {
        "id": "p00fzl8d",
        "name": "BBC Radio Scotland (FM)",
        "image": "bbc_radio_scotland_fm_colour",
    },
    {
        "id": "p00fzl8g",
        "name": "BBC Radio Scotland (MW)",
        "image": "bbc_radio_scotland_colour",
    },
    {
        "id": "p00fzl8b",
        "name": "BBC Radio Scotland (Orkney)",
        "image": "bbc_radio_scotland_colour",
    },
    {
        "id": "p00fzl8j",
        "name": "BBC Radio Scotland (Shetland)",
        "image": "bbc_radio_scotland_colour",
    },
    {
        "id": "p00fzl8h",
        "name": "BBC Radio Sheffield",
        "image": "bbc_radio_sheffield_colour",
    },
    {
        "id": "p00fzl8k",
        "name": "BBC Radio Shropshire",
        "image": "bbc_radio_shropshire_colour",
    },
    {"id": "p00fzl8l", "name": "BBC Radio Solent", "image": "bbc_radio_solent_colour"},
    {"id": "p00fzl8n", "name": "BBC Radio Stoke", "image": "bbc_radio_stoke_colour"},
    {
        "id": "p00fzl8p",
        "name": "BBC Radio Suffolk",
        "image": "bbc_radio_suffolk_colour",
    },
    {"id": "p00fzl8w", "name": "BBC Radio Ulster", "image": "bbc_radio_ulster_colour"},
    {
        "id": "p00fzl8y",
        "name": "BBC Radio Wales (FM)",
        "image": "bbc_radio_wales_fm_colour",
    },
    {
        "id": "p00fzl8x",
        "name": "BBC Radio Wales (LW)",
        "image": "bbc_radio_wales_colour",
    },
    {"id": "p00fzl90", "name": "BBC Radio York", "image": "bbc_radio_york_colour"},
    {
        "id": "p00fzl8m",
        "name": "BBC Somerset",
        "image": "bbc_radio_somerset_sound_colour",
    },
    {"id": "p00fzl8q", "name": "BBC Surrey", "image": "bbc_radio_surrey_colour"},
    {"id": "p00fzl8r", "name": "BBC Sussex", "image": "bbc_radio_sussex_colour"},
    {"id": "p00fzl93", "name": "BBC Tees", "image": "bbc_tees_colour"},
    {
        "id": "p00fzl96",
        "name": "BBC Three Counties Radio",
        "image": "bbc_three_counties_radio_colour",
    },
    {"id": "p00fzl8z", "name": "BBC Wiltshire", "image": "bbc_radio_wiltshire_colour"},
    {"id": "p00fzl9f", "name": "BBC WM 95.6", "image": "bbc_wm_colour"},
    {
        "id": "p02zbmb3",
        "name": "BBC World Service",
        "image": "bbc_world_service_colour",
    },
    {"id": "p02jf21y", "name": "CBeebies Radio", "image": "cbeebies_radio_colour"},
]

_g_initok = False
_g_fetchdays = 30


def maybeinit(a={}):
    global _g_initok

    # Do this always
    setidprefix(bbcidprefix)

    if _g_initok:
        return True

    if "UPMPD_HTTPHOSTPORT" not in os.environ:
        raise Exception("No UPMPD_HTTPHOSTPORT in environment")
    httphp = os.environ["UPMPD_HTTPHOSTPORT"]
    if "UPMPD_PATHPREFIX" not in os.environ:
        raise Exception("No UPMPD_PATHPREFIX in environment")
    pathprefix = os.environ["UPMPD_PATHPREFIX"]
    if "UPMPD_CONFIG" not in os.environ:
        raise Exception("No UPMPD_CONFIG in environment")

    global _g_fetchdays
    val = getOptionValue("bbcprogrammedays")
    if val:
        _g_fetchdays = int(val)

    setMimeAndSamplerate("audio/mpeg", "44100")

    _g_initok = True


# Download the source HTML for the page using requests and parse the page using BeautifulSoup
def get_page(url):
    return BeautifulSoup(requests.get(url).text, "html.parser")


# The media URLs we obtain point to 2 levels of m3u playlists, the second of which is an Apple HTTP
# Live Streaming playlist.
# https://developer.apple.com/library/archive/documentation/NetworkingInternet/Conceptual/StreamingMediaGuide/DeployingHTTPLiveStreaming/DeployingHTTPLiveStreaming.html
# It happens that mpd deals with these fine, not sure that other renderers will. Mitigating this
# would be complicated. The only simple thing we could do is resolve the first level and hope that
# renderers can deal with the Apple part? Else we'd have to proxy the stream.
#
# Also, retrieving the programme details is slow and the listing can timeout when there are many
# entries (in station_date_view()). We'd need to send multiple requests in parallel, either with
# grequests (would need change of programmedetails/station_date_view() interaction), or with
# multiple explicit programmedetails() threads.
def _programmedetails(progid, resdata):
    maybeinit()
    resdata[progid] = None

    try:
        url = "https://www.bbc.co.uk/programmes/" + progid + ".json"
        programme = requests.get(url, timeout=5)
    except:
        uplog("Timed out for %s" % url)
    programme_json = programme.json()["programme"]

    picked_url = None
    metadata = {}
    for version in programme_json["versions"]:
        url = (
            "https://open.live.bbc.co.uk/mediaselector/6/select/version/2.0/mediaset/iptv-all/vpid/"
            + version["pid"]
            + "/format/json"
        )
        uplog("DETAILS: getting: %s" % url, level=4)
        playlist = requests.get(url)
        playlist_json = playlist.json()

        if "media" not in playlist_json:
            # TODO
            continue

        # Filter by only audio items, and order with the highest bitrate first
        audio_items = [
            item for item in playlist_json["media"] if item["kind"] == "audio"
        ]
        audio_items.sort(key=lambda x: x["bitrate"], reverse=True)

        uplog(
            "Found {0} audio items for the programme version {1}".format(
                len(audio_items), version["pid"]
            ),
            level=4,
        )

        # Pick the first stream available for the highest bitrate item
        picked_stream = audio_items[0]
        picked_url = picked_stream["connection"][1]["href"]

        uplog(
            "Picked {0} stream with bitrate {1}".format(
                picked_stream["encoding"], picked_stream["bitrate"]
            ),
            level=4,
        )

        encoding = "audio/mpeg"
        if picked_stream["encoding"] == "aac":
            encoding = "audio/aac"
        # broadcast_date = dateutil.parser.parse(programme_json['first_broadcast_date'])
        # dte = broadcast_date.strftime('%Y-%m-%d, %H:%M') + " "
        dte = ""
        metadata = {
            "url": picked_url,
            "encoding": encoding,
            "bitrate": picked_stream["bitrate"],
            "thumb": "https://ichef.bbci.co.uk/images/ic/480xn/"
            + programme_json["image"]["pid"]
            + ".jpg",
            "icon": "https://ichef.bbci.co.uk/images/ic/480xn/"
            + programme_json["image"]["pid"]
            + ".jpg",
            "title": programme_json["display_title"]["title"],
            "artist": dte + programme_json["display_title"]["subtitle"],
            # Album is the station name.
            "album": programme_json["ownership"]["service"]["title"],
            "comment": programme_json["short_synopsis"],
        }
        # uplog("Metadata %s" % metadata, level=5)
        break

    if picked_url is None:
        uplog("Episode not available to stream")
    else:
        resdata[progid] = metadata


# Parallel fetching of programme data. Threads return their result by setting a dictionary entry
_nthreads = 20


def _fetchdetails(resdata, reqcount):
    thrcount = 0
    ths = []
    datacount = 0
    nthreads = _nthreads
    for progid in resdata.keys():
        th = threading.Thread(target=_programmedetails, args=(progid, resdata))
        th.start()
        ths.append(th)
        thrcount += 1
        if thrcount == nthreads:
            thrcount = 0
            for th in ths:
                th.join()
        datacount = len([k for k in resdata.keys() if resdata[k] is not None])
        if nthreads > reqcount - datacount:
            nthreads = reqcount - datacount
        if datacount >= reqcount:
            break
    for th in ths:
        th.join()


@plugin.route("/")
def root():
    xbmcplugin.add_directory("Stations", root_stations)
    if hasfeedparser:
        xbmcplugin.add_directory("Podcasts", root_podcasts)


@plugin.route("/root_stations")
def root_stations():
    items = []
    for station in stations:
        kwargs = {
            "id": station["id"],
            "name": station["name"],
            #'image': station['image']
        }
        items.append(Playlist(**kwargs))
    xbmcplugin.view(items, xbmcplugin.urls_from_id(station_view, items))


@plugin.route("/station/<station_id>")
def station_view(station_id):
    base = datetime.datetime.today()

    # Create a range of the last 30 days
    for delta in range(_g_fetchdays):
        date = base - datetime.timedelta(days=delta)

        year = "%04d" % date.year
        month = "%02d" % date.month
        day = "%02d" % date.day
        title = date.strftime("%a, %-d %b")
        xbmcplugin.add_directory(
            title,
            plugin.url_for(
                station_date_view,
                station_id=station_id,
                year=year,
                month=month,
                day=day,
            ),
        )


@plugin.route("/station/<station_id>/<year>/<month>/<day>")
def station_date_view(station_id, year, month, day):
    # Load the schedules for the station. Input date is like 2021/11/08
    url = (f"https://www.bbc.co.uk/schedules/{station_id}/{year}/{month}/{day}")
    uplog("station_date_view: fetching %s" % url, level=5)
    schedule = get_page(url)

    result = None

    for tag in schedule.find_all("script", type="application/ld+json"):
        if "RadioEpisode" in tag.contents[0]:
            result = json.loads(tag.contents[0])

    if result is None:
        uplog("Something went wrong parsing the station's schedule")
        return

    uplog("station_date_view: JSON: %s" % result, level=6)
    if _g_do_slices:
        reqcount = xbmcplugin.count
        reqoffs = xbmcplugin.offset
        xbmcplugin.total = len(result["@graph"])
    else:
        reqcount = 1000
        reqoffs = 0

    resdata = {}
    offset = 0
    for episode in result["@graph"]:
        if offset < reqoffs:
            offset += 1
            continue
        progid = episode["identifier"]
        resdata[progid] = None
    _fetchdetails(resdata, reqcount)

    offset = 0
    tracks = []
    trackno = 1 + reqoffs
    for episode in result["@graph"]:
        if offset < reqoffs:
            offset += 1
            continue
        progid = episode["identifier"]
        if not progid in resdata or resdata[progid] is None:
            continue
        trackdata = resdata[progid]

        # We ignore the details title as the one in the episode data is fine.
        date = dateutil.parser.parse(episode["publication"]["startDate"])
        time = date.strftime("%H:%M")
        if "partOfSeries" in episode:
            title = (
                time + ": " + episode["partOfSeries"]["name"] + " - " + episode["name"]
            )
        else:
            title = time + ": " + episode["name"]

        track = {
            "pid": xbmcplugin.objid,
            "id": xbmcplugin.objid + "$" + "%s" % progid,
            "tt": title,
            "dc:title": title,
            "uri": trackdata["url"],
            "tp": "it",
            "upnp:albumArtURI": trackdata["thumb"],
            "upnp:originalTrackNumber": str(trackno),
            # artist has the subtitle, which is already in the title obtained from the episode
            #            'upnp:artist' : trackdata['artist'],
            "duration": "0",
            "upnp:class": "object.item.audioItem.musicTrack",
            "res:mime": trackdata["encoding"],
            "res:samplefreq": "48000",
            "res:bitrate": trackdata["bitrate"],
        }
        xbmcplugin.entries.append(track)
        trackno += 1
        if len(xbmcplugin.entries) >= reqcount:
            break
    uplog("station_date_view: %d entries in result" % len(xbmcplugin.entries), level=5)


@plugin.route("/root_podcasts")
def root_podcasts():
    podcasts = requests.get("https://www.bbc.co.uk/podcasts.json")
    podcasts_json = podcasts.json()["podcasts"]

    # Sort the podcasts by title
    podcasts_ordered = sorted(podcasts_json, key=lambda x: x["title"])

    for podcast in podcasts_ordered:
        arturi = None
        if "imageUrl" in podcast:
            arturi = podcast["imageUrl"].replace("{recipe}", "624x624")
            xbmcplugin.add_directory(
                podcast["title"],
                plugin.url_for(podcast_view, podcast_id=podcast["shortTitle"]),
                arturi=arturi,
            )


@plugin.route("/podcast/<podcast_id>")
def podcast_view(podcast_id):
    podcast = feedparser.parse(
        "https://podcasts.files.bbci.co.uk/" + podcast_id + ".rss"
    )

    image_url = ""
    if "image" in podcast.feed:
        image_url = podcast.feed.image.url

    if _g_do_slices:
        reqcount = xbmcplugin.count
        reqoffs = xbmcplugin.offset
        xbmcplugin.total = len(podcast.entries)
    else:
        reqoffs = 0
        reqcount = 1000

    resdata = {}
    offset = 0
    for entry in podcast.entries:
        if offset < reqoffs:
            offset += 1
            continue
        entry_pid = entry.ppg_canonical.split("/")
        if len(entry_pid) <= 2:
            uplog("No pid could be found for the item at " + entry.link)
            continue
        progid = entry_pid[2]
        resdata[progid] = None
    _fetchdetails(resdata, reqcount)

    trackno = 1 + reqoffs
    offset = 0
    for entry in podcast.entries:
        if offset < reqoffs:
            offset += 1
            continue
        entry_pid = entry.ppg_canonical.split("/")
        if len(entry_pid) <= 2:
            uplog("bad entry_pid for offset %d : %s" % (offset, entry_pid))
            continue
        progid = entry_pid[2]
        if not progid in resdata or resdata[progid] is None:
            uplog("No data for progid %s" % progid)
            continue
        trackdata = resdata[progid]
        entry_date = datetime.datetime.fromtimestamp(
            mktime(entry.published_parsed)
        ).strftime("%Y-%m-%d")
        entry_title = entry_date + ": " + entry.title
        track = {
            "pid": xbmcplugin.objid,
            "id": xbmcplugin.objid + "$" + "%s" % progid,
            "tt": entry_title,
            "uri": trackdata["url"],
            "tp": "it",
            "upnp:albumArtURI": image_url,
            "upnp:originalTrackNumber": str(trackno),
            "dc:title": entry_title,
            "duration": "0",
            "upnp:class": "object.item.audioItem.musicTrack",
            "res:mime": trackdata["encoding"],
            "res:samplefreq": "48000",
            "res:bitrate": trackdata["bitrate"],
        }
        xbmcplugin.entries.append(track)
        trackno += 1
        if len(xbmcplugin.entries) >= reqcount:
            break


@dispatcher.record("browse")
def browse(a):
    global xbmcplugin
    msgproc.log("browse: [%s]" % a)
    if "objid" not in a:
        raise Exception("No objid in args")
    objid = a["objid"]
    if re.match(r"0\$bbc\$", objid) is None:
        raise Exception("bad objid [%s]" % objid)
    maybeinit()

    xbmcplugin = XbmcPlugin(
        bbcidprefix,
        objid,
        routeplugin=plugin,
        offset=int(a["offset"]) if "offset" in a else 0,
        count=int(a["count"]) if "count" in a else 0,
    )

    idpath = objid.replace(bbcidprefix, "", 1)
    plugin.run([idpath])
    # msgproc.log("%s" % xbmcplugin.entries)
    encoded = json.dumps(xbmcplugin.entries)
    ret = {"entries": encoded}
    if _g_do_slices:
        if xbmcplugin.total:
            ret["total"] = str(xbmcplugin.total)
            ret["offset"] = str(xbmcplugin.offset)
    return ret


@dispatcher.record("search")
def search(a):
    msgproc.log("search: [%s]" % a)
    global xbmcplugin
    xbmcplugin = XbmcPlugin(bbcidprefix, "", routeplugin=plugin)
    encoded = json.dumps(xbmcplugin.entries)
    return {"entries": encoded}


msgproc.log("BBC running")
msgproc.mainloop()



# Copyright (C) 2026 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from enum import Enum


class _AlbumPropertyKeyData:

    def __init__(
            self,
            property_key: str = None,
            display_value: str = None,
            unique_value: bool = False,
            max_items: int = 100):
        self.__property_key: str = property_key
        self.__display_value: str = display_value
        self.__unique_value: bool = unique_value
        self.__max_items: int = max_items

    @property
    def property_key(self) -> str:
        return self.__property_key

    @property
    def display_value(self) -> str:
        return self.__display_value

    @property
    def unique_value(self) -> bool:
        return self.__unique_value

    @property
    def max_items(self) -> int:
        return self.__max_items


class AlbumPropertyKey(Enum):

    GENRE = _AlbumPropertyKeyData(display_value="Genre")
    MOOD = _AlbumPropertyKeyData(display_value="Mood")
    LABEL = _AlbumPropertyKeyData(display_value="Label")
    DECADE = _AlbumPropertyKeyData(display_value="Decade", unique_value=True)
    YEAR = _AlbumPropertyKeyData(display_value="Year", unique_value=True)
    ARTIST = _AlbumPropertyKeyData(display_value="Artist")
    ARTIST_CONTRIBUTOR = _AlbumPropertyKeyData(display_value="Artist/Contributor")
    CONTRIBUTOR = _AlbumPropertyKeyData(display_value="Contributor")
    RELEASE_TYPE = _AlbumPropertyKeyData(display_value="Release Type")
    LOSSLESS_STATUS = _AlbumPropertyKeyData(display_value="Compression Type", unique_value=True)
    QUALITY_BADGE = _AlbumPropertyKeyData(display_value="Quality Badge", unique_value=True)
    BIT_DEPTH = _AlbumPropertyKeyData(display_value="Bit Depth")
    SAMPLING_RATE = _AlbumPropertyKeyData(display_value="Sampling Rate")
    CHANNEL_COUNT = _AlbumPropertyKeyData(display_value="Channel Count")

    @property
    def property_key(self) -> str:
        return self.value.property_key if self.value.property_key else self.name.lower()

    @property
    def display_value(self) -> str:
        return self.value.display_value if self.value.display_value else self.name

    @property
    def unique_value(self) -> bool:
        return self.value.unique_value

    @property
    def max_items(self) -> int:
        return self.value.max_items


def get_album_property_key(property_key: str) -> AlbumPropertyKey:
    apk: AlbumPropertyKey
    for apk in AlbumPropertyKey:
        if property_key == apk.property_key:
            return apk
    # no match
    return None


class AlbumPropertyKeyValue:

    def __init__(
            self,
            key: str,
            value: str):
        self.__key: str = key
        self.__value: str = value

    @property
    def key(self) -> str:
        return self.__key

    @property
    def value(self) -> str:
        return self.__value



# Copyright (C) 2023 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Callable
from tidalapi.album import Album as TidalAlbum


class AlbumSortCriteria:

    def __init__(self, extractor: Callable[[TidalAlbum], any], comparator: Callable[[any, any], int]):
        self._extractor: Callable[[TidalAlbum], any] = extractor
        self._comparator: Callable[[any, any], int] = comparator

    def compare(self, left: TidalAlbum, right: TidalAlbum) -> int:
        return self._comparator(self._extractor(left), self._extractor(right))



#################################
# Copyright (C) 2016 J.F.Dockes
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the
#   Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
########################################################
# Command communication module and utilities. See commands in cmdtalk.h
#
# All data is binary. This is important for Python3
# All parameter names are converted to and processed as str/unicode

import sys
import os
import traceback
import signal
from typing import Any, Union, TextIO, BinaryIO, Text, AnyStr, Callable, Optional, Dict, Tuple


def makebytes(data: AnyStr) -> bytes:
    if data is None:
        return b""
    if isinstance(data, str):
        return data.encode("UTF-8")
    else:
        return data


def breakwrite(outfile: BinaryIO, data: bytes) -> None:
    if sys.platform != "win32":
        outfile.write(data)
    else:
        # On Windows, writing big chunks can fail with a "not enough space"
        # error. Seems a combined windows/python bug, depending on versions.
        # See https://bugs.python.org/issue11395
        # In any case, just break it up
        total = len(data)
        bs = 4 * 1024
        offset = 0
        while total > 0:
            if total < bs:
                tow = total
            else:
                tow = bs
            outfile.write(data[offset : offset + tow])
            offset += tow
            total -= tow


############################################
# CmdTalk implements the communication protocol with the master
# process. It calls an external method to use the args and produce
# return data.
class CmdTalk(object):

    def __init__(self, outfile: TextIO = sys.stdout, infile: TextIO = sys.stdin,
                 exitfunc: Optional[Callable] = None):
        try:
            self.myname = os.path.basename(sys.argv[0])
        except:
            self.myname = "???"

        self.outfile = outfile
        self.infile = infile
        self.exitfunc = exitfunc

        if sys.platform == "win32":
            import msvcrt
            msvcrt.setmode(self.outfile.fileno(), os.O_BINARY)
            msvcrt.setmode(self.infile.fileno(), os.O_BINARY)

        if not hasattr(self, 'debugfile'):
            self.debugfile = None
        if not hasattr(self, 'nodecodeinput'):
            self.nodecodeinput = False
        self.errfout: TextIO = sys.stderr
        if self.debugfile:
            self.errfout: TextIO = open(self.debugfile, "a")

    def log(self, s: AnyStr, doexit: int = 0, exitvalue: int = 1) -> None:
        print(f"CMDTALK: {self.myname}: {s!r}", file=self.errfout)
        if doexit:
            if self.exitfunc:
                self.exitfunc(exitvalue)
            sys.exit(exitvalue)

    # Read single parameter from process input: line with param name and size
    # followed by data. The param name is returned as str/unicode, the data
    # as bytes or str, depending on the nodecodeinput option.
    def readparam(self) -> Tuple[Text, Union[bytes, str]]:
        inf = self.infile.buffer
        s = inf.readline()
        if s == b"":
            if self.exitfunc:
                self.exitfunc(0)
            # Our father process is probably going to send us a SIGTERM.  On some platforms (BSD,
            # MacOS), an exception is sometimes generated during exit processing, probably randomly
            # depending on where we are in the process when we receive the signal. Ignoring SIGTERM
            # while we exit mostly fixes the problem (we very rarely get a message about a signal
            # race condition). None of this affects the working of the program anyway, just an issue
            # with error messages.
            signal.signal(signal.SIGTERM, signal.SIG_IGN)
            sys.exit(0)

        s = s.rstrip(b"\n")

        if s == b"":
            return ("", b"")
        l = s.split()
        if len(l) != 2:
            self.log(b"bad line: [" + s + b"]", 1, 1)

        paramname = l[0].decode("ASCII").rstrip(":")
        paramsize = int(l[1])
        if paramsize > 0:
            paramdata = inf.read(paramsize)
            if len(paramdata) != paramsize:
                self.log(
                    "Bad read: wanted %d, got %d" % (paramsize, len(paramdata)), 1, 1
                )
        else:
            paramdata = b""
        if not self.nodecodeinput:
            try:
                paramstr: str = paramdata.decode("utf-8")
                return (paramname, paramstr)
            except Exception as ex:
                self.log("Exception decoding param: %s" % ex)
                paramdata = b""

        # self.log("paramname [%s] paramsize %d value [%s]" %
        #          (paramname, paramsize, paramdata))
        return (paramname, paramdata)

    def senditem(self, nm: Text, _data: AnyStr) -> None:
        data: bytes = makebytes(_data)
        l = len(data)
        self.outfile.buffer.write(makebytes("%s: %d\n" % (nm, l)))
        breakwrite(self.outfile.buffer, data)

    # Send answer
    def answer(self, outfields: Dict[Text, AnyStr]) -> None:
        for nm, value in outfields.items():
            # self.log("Senditem: [%s] -> [%s]" % (nm, value))
            self.senditem(nm, value)

        # End of message: empty line
        print(file=self.outfile)
        self.outfile.flush()
        # self.log("done writing data")

    # Call processor with input params, send result. This base version works with, for example
    # the cmdtalkplugin processor.
    def processmessage(self, processor, params: Dict[Text, Union[bytes, str]]) -> None:
        # In normal usage we try to recover from processor errors, but
        # we sometimes want to see the real stack trace when testing
        safeexec = True
        if safeexec:
            try:
                outfields = processor.process(params)
            except Exception as err:
                self.log("processmessage: processor raised: [%s]" % err)
                traceback.print_exc()
                outfields = {}
                outfields["cmdtalkstatus"] = "1"
                outfields["cmdtalkerrstr"] = str(err)
        else:
            outfields = processor.process(params)

        self.answer(outfields)

    # Loop on messages from our master
    def mainloop(self, processor: Any) -> None:
        while 1:
            # self.log("waiting for command")
            params = dict()

            # Read at most 20 parameters (normally 1 or 2), stop at empty line
            # End of message is signalled by empty paramname
            for i in range(20):
                paramname, paramdata = self.readparam()
                if paramname == "":
                    break
                params[paramname] = paramdata

            # Got message, act on it
            self.processmessage(processor, params)


# Common main routine for testing: either run the normal protocol
# engine or a local loop. This means that you can call
# cmdtalk.main(proto,processor) instead of proto.mainloop(processor)
# from your module, and get the benefits of command line testing
def main(proto: CmdTalk, processor: Any) -> None:
    if len(sys.argv) == 1:
        proto.mainloop(processor)
        # mainloop does not return. Just in case
        sys.exit(1)

    # Not running the main loop: run one processor call for debugging
    def usage():
        print("Usage: cmdtalk.py pname pvalue [pname pvalue...]", file=sys.stderr)
        sys.exit(1)

    def debprint(out, s):
        breakwrite(out, makebytes(s + "\n"))

    args = sys.argv[1:]
    if len(args) == 0 or len(args) % 2 != 0:
        usage()
    params = dict()
    for i in range(int(len(args) / 2)):
        params[args[2 * i]] = args[2 * i + 1]
    res = processor.process(params)

    ioout = sys.stdout.buffer

    for nm, value in res.items():
        # debprint(f"Senditem: [{nm}] -> [{value}]")
        bdata = makebytes(value)
        debprint(ioout, "%s->" % nm)
        breakwrite(ioout, bdata)
        ioout.write(b"\n")



# Copyright (C) 2024,2025 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import upmplgutils
import constants
from tidalapi import Quality as TidalQuality


def get_plugin_config_variable_name(name: str) -> str:
    return f"{constants.PluginConstant.PLUGIN_NAME.value}{name}"


def get_option_value_as_bool(nm: str, default_value: int) -> bool:
    return get_option_value(nm, default_value) == 1


def get_option_value(nm, dflt: any = None):
    return upmplgutils.getOptionValue(get_plugin_config_variable_name(nm), dflt)


def get_config_param_as_str(configuration_parameter: constants.ConfigParam) -> str:
    dv: str | None = configuration_parameter.default_value
    if dv is not None and not isinstance(dv, str):
        raise Exception(f"Invalid default value for [{configuration_parameter.key}]")
    v: any = get_option_value(configuration_parameter.key, dv)
    if v is None:
        return None
    # v is set, check type!
    if isinstance(v, str):
        return v
    # try to convert to string
    return str(v)


def get_config_param_as_int(configuration_parameter: constants.ConfigParam) -> str:
    dv: int | None = configuration_parameter.default_value
    if dv is not None and not isinstance(dv, int):
        raise Exception(f"Invalid default value for [{configuration_parameter.key}]")
    v: any = get_option_value(configuration_parameter.key, dv)
    if v is None:
        return None
    # v is set, check type!
    if isinstance(v, int):
        return v
    # try to convert to int
    return int(v)


def get_config_param_as_bool(configuration_parameter: constants.ConfigParam) -> bool:
    default_value_as_int: int = 0
    dv: any = configuration_parameter.default_value
    if isinstance(dv, int):
        default_value_as_int = 1 if dv == 1 else 0
    elif isinstance(dv, bool):
        default_value_as_int = 1 if dv else 0
    return get_option_value_as_bool(
        configuration_parameter.key,
        default_value_as_int)


def __getPluginOptionAsBool(plugin_param_name: str, default_value: bool | int = None) -> bool:
    sanitized_default_value: int | bool = 0 if not default_value else default_value
    if type(sanitized_default_value) is bool:
        sanitized_default_value = 1 if default_value else 0
    elif type(sanitized_default_value) is int:
        if sanitized_default_value not in [0, 1]:
            raise Exception(f"Invalid boolean value for [{plugin_param_name}], should be 0 or 1")
    return (getPluginOptionValue(plugin_param_name, sanitized_default_value) == 1)


def getPluginOptionValue(option_key: str, dflt=None):
    return upmplgutils.getOptionValue(
        nm=f"{constants.PluginConstant.PLUGIN_NAME.value}{option_key}",
        dflt=dflt)


def getWebServerDocumentRoot() -> str:
    # return getUpnpWebDocRoot(constants.plugin_name)
    return upmplgutils.getOptionValue("webserverdocumentroot")


enable_read_stream_metadata: bool = get_config_param_as_bool(constants.ConfigParam.ENABLE_READ_STREAM_METADATA)

enable_assume_bitdepth: bool = __getPluginOptionAsBool(
    "enableassumebitdepth",
    constants.default_enable_assume_bitdepth)


tracks_per_page: int = getPluginOptionValue(
    "tracksperpage",
    constants.default_tracks_per_page)


mix_items_per_page: int = getPluginOptionValue(
    "mixitemsperpage",
    constants.default_mix_items_per_page)


playlist_items_per_page: int = getPluginOptionValue(
    "playlistitemsperpage",
    constants.default_playlist_items_per_page)


dump_track_to_entry_result: bool = getPluginOptionValue(
    "dumptracktoentryresult",
    constants.default_dump_track_to_entry_result)


allow_guess_stream_info_from_other_album_track: bool = getPluginOptionValue(
    "allowguessstreaminfofromotheralbumtrack",
    constants.default_allow_guess_stream_info_from_other_album_track)


max_file_age_seconds: int = getPluginOptionValue(
    "maxfileageseconds",
    constants.default_max_file_age_seconds)


max_get_stream_info_mix_or_playlist: bool = getPluginOptionValue(
    "maxgetstreaminfomixorplaylist",
    constants.default_max_get_stream_info_mix_or_playlist)


page_items_per_page: int = getPluginOptionValue(
    "pageitemsperpage",
    constants.default_page_items_per_page)


albums_per_page: int = getPluginOptionValue(
    "albumsperpage",
    constants.default_albums_per_page)


artists_per_page: int = getPluginOptionValue(
    "artistsperpage",
    constants.default_artists_per_page)


show_album_id: bool = getPluginOptionValue(
    "showalbumid",
    constants.default_show_album_id)


badge_favorite_album: bool = getPluginOptionValue(
    "badgefavoritealbum",
    constants.default_badge_favorite_album)


titleless_single_album_view: bool = getPluginOptionValue(
    "titlelesssinglealbumview",
    constants.default_titleless_single_album_view)


skip_non_stereo: bool = __getPluginOptionAsBool(
    "skipnonstereo",
    constants.default_skip_non_stereo)


log_intermediate_url: bool = __getPluginOptionAsBool(
    "logintermediateurl",
    constants.default_log_intermediate_url)


prepend_number_in_album_list: bool = __getPluginOptionAsBool(
    "prependnumberinitemlist",
    constants.default_prepend_number_in_item_list)


serve_mode: str = getPluginOptionValue(
    "servemode",
    constants.default_serve_mode)


max_playlist_or_mix_items_per_page: int = getPluginOptionValue(
    "maxplaylistitemsperpage",
    constants.default_max_playlist_or_mix_items_per_page)


listen_queue_playlist_name: str = getPluginOptionValue(
    "listenqueueplaylistname",
    constants.default_listen_queue_playlist_name)


display_quality_badge: bool = __getPluginOptionAsBool(
    "displayqualitybadge",
    constants.default_display_quality_badge)


def get_enable_image_caching() -> bool:
    return __getPluginOptionAsBool("enableimagecaching", constants.default_enable_image_caching)


__fallback_for_missing_quality: dict[str, str] = {
    TidalQuality.hi_res_lossless: TidalQuality.high_lossless,
    TidalQuality.high_lossless: TidalQuality.high_lossless,
    TidalQuality.low_320k: TidalQuality.low_320k,
    TidalQuality.low_96k: TidalQuality.low_96k
}


def get_fallback_quality_when_missing() -> str:
    max_audio_quality: str = get_config_param_as_str(constants.ConfigParam.AUDIO_QUALITY)
    if not max_audio_quality or max_audio_quality not in __fallback_for_missing_quality:
        return None
    return __fallback_for_missing_quality[max_audio_quality]


def get_dump_image_caching() -> bool:
    return __getPluginOptionAsBool(
        "dumpimagecaching",
        constants.default_dump_image_caching)


def get_override_country_code() -> str:
    return getPluginOptionValue("overridecountrycode")


def get_tile_image_expiration_time_sec() -> int:
    return getPluginOptionValue(
        "tileimageexpirationtimesec",
        constants.tile_image_expiration_time_sec)


def get_allow_favorite_actions() -> bool:
    return __getPluginOptionAsBool(
        constants.ConfigParam.ALLOW_FAVORITE_ACTIONS.key,
        constants.ConfigParam.ALLOW_FAVORITE_ACTIONS.default_value)


def get_allow_bookmark_actions() -> bool:
    return __getPluginOptionAsBool(
        constants.ConfigParam.ALLOW_BOOKMARK_ACTIONS.key,
        constants.ConfigParam.ALLOW_BOOKMARK_ACTIONS.default_value)


def get_allow_statistics_actions() -> bool:
    return __getPluginOptionAsBool(
        constants.ConfigParam.ALLOW_STATISTICS_ACTIONS.key,
        constants.ConfigParam.ALLOW_STATISTICS_ACTIONS.default_value)



# Copyright (C) 2023,2024 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from radio_station_entry import RadioStationEntry

codec_flac: str = "flac"
codec_aac: str = "aac"
codec_mp3: str = "mp3"

mimetype_flac: str = "audio/flac"
mimetype_aac: str = "audio/aac"
mimetype_mp3: str = "audio/mpeg"

title_mother_earth_radio : str = "Mother Earth Radio"
title_mother_earth_klassik : str = "Mother Earth Klassik"
title_mother_earth_instrumental : str = "Mother Earth Instrumental"
title_mother_earth_jazz : str = "Mother Earth Jazz"

radio_station_list: list[RadioStationEntry] = list()

def __add(
        codec: str,
        url: str,
        title: str,
        mimetype: str,
        bit_depth : int,
        sampling_rate : int,
        channel_count : int = 2):
    id: int = len(radio_station_list) + 1
    e: RadioStationEntry = RadioStationEntry(
        id=id,
        url=url,
        codec=codec,
        title=title,
        mimetype=mimetype,
        bit_depth=bit_depth,
        sampling_rate=sampling_rate,
        channel_count=channel_count
    )
    radio_station_list.append(e)


__add(
    url="https://motherearth.streamserver24.com/listen/motherearth/motherearth",
    title=title_mother_earth_radio,
    codec=codec_flac,
    mimetype=mimetype_flac,
    bit_depth=24,
    sampling_rate=192000,
)
__add(
    url="https://motherearth.streamserver24.com/listen/motherearth/motherearth.aac",
    title=title_mother_earth_radio,
    codec=codec_aac,
    mimetype=mimetype_aac,
    bit_depth=24,
    sampling_rate=96000,
)
__add(
    url="https://motherearth.streamserver24.com/listen/motherearth/motherearth.mp3",
    title=title_mother_earth_radio,
    codec=codec_mp3,
    mimetype=mimetype_mp3,
    bit_depth=16,
    sampling_rate=44100,
)
__add(
    url="https://motherearth.streamserver24.com/listen/motherearth_klassik/motherearth.klassik",
    title=title_mother_earth_klassik,
    codec=codec_flac,
    mimetype=mimetype_flac,
    bit_depth=24,
    sampling_rate=192000,
)
__add(
    url="https://motherearth.streamserver24.com/listen/motherearth_klassik/motherearth.klassik.aac",
    title=title_mother_earth_klassik,
    codec=codec_aac,
    mimetype=mimetype_aac,
    bit_depth=24,
    sampling_rate=96000,
)
__add(
    url="https://motherearth.streamserver24.com/listen/motherearth_klassik/motherearth.klassik.mp3",
    title=title_mother_earth_klassik,
    codec=codec_mp3,
    mimetype=mimetype_mp3,
    bit_depth=16,
    sampling_rate=44100,
)
__add(
    url="https://motherearth.streamserver24.com/listen/motherearth_instrumental/motherearth.instrumental",
    title=title_mother_earth_instrumental,
    codec=codec_flac,
    mimetype=mimetype_flac,
    bit_depth=24,
    sampling_rate=192000,
)
__add(
    url="https://motherearth.streamserver24.com/listen/motherearth_instrumental/motherearth.instrumental.aac",
    title=title_mother_earth_instrumental,
    codec=codec_aac,
    mimetype=mimetype_aac,
    bit_depth=24,
    sampling_rate=96000,
)
__add(
    url="https://motherearth.streamserver24.com/listen/motherearth_instrumental/motherearth.instrumental.mp3",
    title=title_mother_earth_instrumental,
    codec=codec_mp3,
    mimetype=mimetype_mp3,
    bit_depth=16,
    sampling_rate=44100,
)
__add(
    url="https://motherearth.streamserver24.com/listen/motherearth_jazz/motherearth.jazz",
    title=title_mother_earth_jazz,
    codec=codec_flac,
    mimetype=mimetype_flac,
    bit_depth=24,
    sampling_rate=192000,
)
__add(
    url="https://motherearth.streamserver24.com/listen/motherearth_jazz/motherearth.jazz.mp4",
    title=title_mother_earth_jazz,
    codec=codec_aac,
    mimetype=mimetype_aac,
    bit_depth=24,
    sampling_rate=96000,
)
__add(
    url="https://motherearth.streamserver24.com/listen/motherearth_jazz/motherearth.jazz.mp3",
    title=title_mother_earth_jazz,
    codec=codec_mp3,
    mimetype=mimetype_mp3,
    bit_depth=16,
    sampling_rate=44100,
)



# Copyright (C) 2026 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from enum import Enum
from keyvaluecaching import KeyValueTableName


class DeletedTableName(Enum):

    ALBUM_COVER_ART_BY_ARTIST_V1 = "album_cover_art_by_artist_v1"


class TableName(Enum):
    ALBUM_METADATA_V1 = "album_metadata_v1"
    ARTIST_METADATA_V1 = "artist_metadata_v1"
    ARTIST_ROLE_V1 = "artist_role_v1"
    ALBUM_ARTIST_V1 = "album_artist_v1"
    ALBUM_DISC_V1 = "album_disc_v1"
    ALBUM_GENRE_V1 = "album_genre_v1"
    ALBUM_RECORD_LABEL_V1 = "album_record_label_v1"
    ALBUM_MOOD_V1 = "album_mood_v1"
    ALBUM_RELEASE_TYPE_V1 = "album_release_type_v1"
    SONG_METADATA_V1 = "song_metadata_v1"
    KV_CACHE_V1 = KeyValueTableName.TABLE_NAME_V1.table_name
    SONG_ALBUM_ARTIST_V1 = "song_album_artist_v1"
    SONG_ARTIST_V1 = "song_artist_v1"
    SONG_CONTRIBUTOR_V1 = "song_contributor_v1"
    ALBUM_PROPERTY_V1 = "album_property_v1"
    DB_VERSION = "db_version"



#!/usr/bin/env python3
#
# A lot of code copied from the Kodi Tidal addon which is:
# Copyright (C) 2014 Thomas Amland
#
# Additional code and modifications:
# Copyright (C) 2020 J.F.Dockes
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import os
import json
import re
import base64

import cmdtalkplugin
from upmplgutils import *
from xbmcplug import *

# Using tidal Kodi addon routing module
from routing import Plugin

plugin = Plugin("")

import session

_rootid = "0$hra$"
_rootidre = r"0\$hra\$"

# Func name to method mapper
dispatcher = cmdtalkplugin.Dispatch()
# Pipe message handler
msgproc = cmdtalkplugin.Processor(dispatcher)

sess = session.Session()


def maybelogin():
    if sess.isloggedin():
        return True

    global httphp
    if "UPMPD_HTTPHOSTPORT" not in os.environ:
        raise Exception("No UPMPD_HTTPHOSTPORT in environment")
    httphp = os.environ["UPMPD_HTTPHOSTPORT"]
    global pathprefix
    if "UPMPD_PATHPREFIX" not in os.environ:
        raise Exception("No UPMPD_PATHPREFIX in environment")
    pathprefix = os.environ["UPMPD_PATHPREFIX"]
    if "UPMPD_CONFIG" not in os.environ:
        raise Exception("No UPMPD_CONFIG in environment")

    username, password = getserviceuserpass("hra")
    if not username or not password:
        raise Exception("hrauser and/or hrapass not set in configuration")
    lang = getOptionValue("hralang")
    if not lang:
        lang = "en"
    setMimeAndSamplerate("application/flac", "44100")
    return sess.login(username, password, lang)


@dispatcher.record("trackuri")
def trackuri(a):
    global pathprefix
    msgproc.log("trackuri: [%s]" % a)
    trackid = trackid_from_urlpath(pathprefix, a)
    maybelogin()
    url = sess.get_media_url(trackid) or ""
    # msgproc.log("%s" % media_url)
    return {"media_url": url}


def track_list(tracks):
    xbmcplugin.entries += trackentries(httphp, pathprefix, xbmcplugin.objid, tracks)


@dispatcher.record("browse")
def browse(a):
    global xbmcplugin
    uplog("browse: [%s]" % a)
    if "objid" not in a:
        raise Exception("No objid in args")
    objid = a["objid"]
    bflg = a["flag"] if "flag" in a else "children"

    if re.match(_rootidre, objid) is None:
        raise Exception("bad objid [%s]" % objid)

    xbmcplugin = XbmcPlugin(_rootid, objid, routeplugin=plugin)
    maybelogin()

    idpath = objid.replace(_rootid, "", 1)
    if bflg == "meta":
        m = re.match(r".*\$(.+)$", idpath)
        if m:
            trackid = m.group(1)
            track = sess.get_track(trackid)
            track_list([track])
    else:
        plugin.run([idpath])
    encoded = json.dumps(xbmcplugin.entries)
    return {"entries": encoded}


@plugin.route("/")
def root():
    if not maybelogin():
        xbmcplugin.add_directory("LOGIN FAILED", "/")
        return
    xbmcplugin.add_directory("Categories", allCategories)
    xbmcplugin.add_directory("Genres", allGenres)
    # These seem mostly empty. The hra app uses alleditor
    #    xbmcplugin.add_directory('Editor playlists: Moods', allEditorMoods)
    #    xbmcplugin.add_directory('Editor playlists: Genres', allEditorGenres)
    #    xbmcplugin.add_directory('Editor playlists: Themes', allEditorThemes)
    xbmcplugin.add_directory("Editor Playlists", allEditorPlaylists)
    xbmcplugin.add_directory("My Playlists", allUserPlaylists)
    xbmcplugin.add_directory("My Albums", allUserAlbums)
    xbmcplugin.add_directory("My Tracks", allUserTracks)


@plugin.route("/allcategories")
def allCategories():
    items = sess.get_categories()
    xbmcplugin.view(items, xbmcplugin.urls_from_id(category_view, items))


@plugin.route("/allgenres")
def allGenres():
    items = sess.get_genres()
    xbmcplugin.view(items, xbmcplugin.urls_from_id(category_view, items))


@plugin.route("/category/<catg_id>")
def category_view(catg_id):
    decoded = session.decode_prefix(catg_id)
    uplog("category_view: id [%s] decoded [%s]" % (catg_id, decoded))
    items = sess.get_catg_albums(decoded)
    xbmcplugin.view(items, xbmcplugin.urls_from_id(album_view, items))


@plugin.route("/alleditorgenres")
def allEditorGenres():
    items = sess.get_editorgenres()
    xbmcplugin.view(items, xbmcplugin.urls_from_id(edplaylist_view, items))


@plugin.route("/alleditormoods")
def allEditorMoods():
    items = sess.get_editormoods()
    xbmcplugin.view(items, xbmcplugin.urls_from_id(edplaylist_view, items))


@plugin.route("/alleditorthemes")
def allEditorThemes():
    items = sess.get_editorthemes()
    xbmcplugin.view(items, xbmcplugin.urls_from_id(edplaylist_view, items))


@plugin.route("/alleditorplaylists")
def allEditorPlaylists():
    items = sess.get_alleditorplaylists()
    xbmcplugin.view(items, xbmcplugin.urls_from_id(edplaylist_view, items))


@plugin.route("/editorplaylist/<id>")
def edplaylist_view(id):
    track_list(sess.get_editor_playlist(id))


@plugin.route("/alluserplaylists")
def allUserPlaylists():
    items = sess.get_alluserplaylists()
    xbmcplugin.view(items, xbmcplugin.urls_from_id(userplaylist_view, items))


@plugin.route("/alluseralbums")
def allUserAlbums():
    items = sess.get_alluseralbums()
    xbmcplugin.view(items, xbmcplugin.urls_from_id(album_view, items))


@plugin.route("/allusertracks")
def allUserTracks():
    track_list(sess.get_allusertracks())


@plugin.route("/userplaylist/<id>")
def userplaylist_view(id):
    track_list(sess.get_user_playlist(id))


@plugin.route("/album/<album_id>")
def album_view(album_id):
    track_list(sess.get_album_tracks(album_id))


@dispatcher.record("search")
def search(a):
    global xbmcplugin

    msgproc.log("search: %s" % a)

    objid = a["objid"]
    field = a["field"] if "field" in a else None
    value = a["value"]
    objkind = a["objkind"] if "objkind" in a and a["objkind"] else None

    if re.match(_rootidre, objid) is None:
        raise Exception("bad objid [%s]" % objid)

    xbmcplugin = XbmcPlugin(_rootid, objid, routeplugin=plugin)
    maybelogin()

    idpath = objid.replace(_rootid, "", 1)

    catg = ""
    # Both genres and categories have an objid path beginning with category
    if idpath.startswith("/category"):
        catg = session.decode_prefix(idpath.replace("/category/", "", 1))

    # HRA search always returns albums anyway.
    if catg:
        albums = sess.searchCategory(value, catg)
    else:
        albums = sess.search(value)
    xbmcplugin.view(albums, xbmcplugin.urls_from_id(album_view, albums))
    # msgproc.log("%s" % xbmcplugin.entries)
    encoded = json.dumps(xbmcplugin.entries)
    return {"entries": encoded}


msgproc.log("HighresAudio running")
msgproc.mainloop()



# Copyright (C) 2017-2034 J.F.Dockes
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the
#   Free Software Foundation, Inc.,
#   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#

import os
import time
import bottle
import mutagen
import re

from upmplgutils import uplog
from uprclutils import embedded_open
import uprclinit

# DNS rebinding mitigation. We check that the HOST header is either the configured host:port (with a
# possible domain-based host) or a numeric address.
_hostre = re.compile(r"[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+:[0-9]+|\[[0-9A-Fa-f:]+\]:[0-9]+")


def _checkhost():
    httphp = uprclinit.getHttphp()
    if "host" in bottle.request.headers:
        host = bottle.request.headers["host"]
        if host == httphp:
            return True
        if not _hostre.match(host):
            uplog(f"Streamer: Bad Host <{host}> (httphp: {httphp})")
            return False
    return True


@bottle.route("/")
@bottle.post("/")
@bottle.view("main")
def main():
    if not _checkhost():
        return bottle.HTTPResponse(status=404)
    # Using params.get will work with either the form's POST or an URL GET query value
    # See https://bottlepy.org/docs/dev/tutorial.html#request-data
    what = bottle.request.params.get("what")
    #uplog("bottle:main: what value is %s" % what)

    status = uprclinit.updaterunning()
    if not status:
        status = "Ready"

    reloadsecs = "1"
    if what == "Update Index":
        uprclinit.start_index_update()
    elif what == "Reset Index":
        uprclinit.start_index_update(rebuild=True)
    elif what == "Refresh Status":
        reloadsecs = ""
    elif not what:
        if status == "Updating":
            reloadsecs = "2"
        elif status == "Rebuilding":
            reloadsecs = "10"
        else:
            reloadsecs = ""

    return {
        "title": status,
        "status": status,
        "reloadsecs": reloadsecs,
        "friendlyname": uprclinit.getFriendlyname(),
    }


@bottle.route("/static/<filepath:path>")
def static(filepath):
    # uplog("control: static: filepath %s datadir %s" % (filepath, datadir))
    if not _checkhost():
        return bottle.HTTPResponse(status=404)
    return bottle.static_file(filepath, root=os.path.join(datadir, "static"))


# Object for streaming data from a given subtree (topdirs entry more
# or less). This is needed just because as far as I can see, a
# callback can't know the route it was called for, so we record it
# when creating the object.
class Streamer(object):
    def __init__(self, root):
        self.root = root.replace("\\:", ":")

    def __call__(self, filepath):
        if not _checkhost():
            return bottle.HTTPResponse(status=404)
        uplog(f"Streamer.__call__: myroot [{self.root}] filepath [{filepath}]", level=4)
        embedded = True if "embed" in bottle.request.query else False
        if embedded:
            # Embedded image urls have had a .jpg or .png
            # appended. Remove it to restore the track path name.
            i = filepath.rfind(".")
            filepath = filepath[:i]
            apath = os.path.join(self.root, filepath)
            ctype, size, f = embedded_open(apath)
            fs = os.stat(apath)
            lm = time.strftime("%a, %d %b %Y %H:%M:%S GMT", time.gmtime(fs.st_mtime))
            bottle.response.set_header("Last-Modified", lm)
            bottle.response.set_header("Content-type", ctype)
            bottle.response.set_header("Content-Length", size)
            return f

        # Binary paths: transmitted as follows (See bottle._handle())
        #   binarypath->binarypath.decode('latin1')->urlquote()->NETWORK->
        #   urlunquote()->encode('latin1').decode('utf-8', 'ignore')
        # If there are non utf-8 characters in the path, the result is
        # missing them, and file access fails. However, we get the
        # binary result from urlunquote)_ in the 'bottle.raw_path'
        # request environment variable, and try to use it if the
        # normal path is not accessible.
        fullpath = os.path.join(self.root, filepath)
        root = "/"
        if not os.path.exists(fullpath):
            fullpath = bottle.request.environ.get("bottle.raw_path")
            fullpath = fullpath.encode("latin1")
            root = b"/"
            if not os.path.exists(fullpath):
                uplog("uprcl: no such file: %s" % fullpath)
                return bottle.HTTPResponse(status=404)

        uplog(f"Streaming: {fullpath}")
        mutf = mutagen.File(fullpath)
        if mutf:
            return bottle.static_file(fullpath, root=root, mimetype=mutf.mime[0])
        else:
            return bottle.static_file(fullpath, root=root)


# Bottle handle both the streaming and control requests.
def runbottle(host="0.0.0.0", port=9278, topdirs="", pathprefix=""):
    global datadir
    uplog(
        f"runbottle: version {bottle.__version__} host {host} port {port} topdirs {topdirs} " \
        f"pathprefix {pathprefix}"
    )
    datadir = os.path.dirname(__file__)
    datadir = os.path.join(datadir, "bottle")
    bottle.TEMPLATE_PATH = (os.path.join(datadir, "views"),)

    # Add routes for all the recoll start directories.
    # As I don't know how a bottle method can retrieve the route it
    # was called from, we create a callable for each prefix.
    for rt in topdirs:
        if rt[-1] != "/":
            rt += "/"
        streamer = Streamer(rt)
        rt = rt.replace(":", "\\:")
        rt += "<filepath:path>"
        uplog(f"runbottle: adding route for: {rt}")
        bottle.route(rt, "GET", streamer)

    bottle.run(server="waitress", host=host, port=port)



# Copyright (C) 2023,2024,2025 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import upmplgmodels
import upmpdmeta
import album_util
import constants

from subsonic_connector.album import Album
import html


def set_entry_value(key_name: str, key_value: str, target: dict):
    if key_name and key_value:
        target[key_name] = key_value


def set_class(upnp_class: str, target: dict):
    target["upnp:class"] = upnp_class


def set_class_artist(target: dict):
    set_class(target=target, upnp_class=upmplgmodels.Artist.upnpclass)


def set_class_music_track(target: dict):
    set_class(target=target, upnp_class=upmplgmodels.Track.upnpclass)


def set_class_album(target: dict):
    set_class(target=target, upnp_class=upmplgmodels.Album.upnpclass)


def set_class_playlist_container(target: dict):
    set_class(target=target, upnp_class=upmplgmodels.Playlist.upnpclass)


def set_track_number(track_number: str, target: dict):
    target["upnp:originalTrackNumber"] = track_number


def get_album_art_uri(entry: dict):
    return entry["upnp:albumArtURI"] if "upnp:albumArtURI" in entry else None


def set_album_art_from_uri(album_art_uri: str, target: dict):
    set_entry_value(key_name="upnp:albumArtURI", key_value=album_art_uri, target=target)


def set_track_title(track_title: str, target: dict):
    set_entry_value(key_name="tt", key_value=track_title, target=target)


def set_album_id(album_id: str, target: dict):
    target["album_id"] = album_id


def set_artist(artist: str, target: dict):
    target["upnp:" + constants.UpnpMeta.ARTIST.value] = artist


def set_date_from_album(album: Album, target: dict):
    date_str: str = album_util.get_formatted_album_original_release_date(album)
    if not date_str:
        # fallback to year
        date_str = str(album.getYear()) if album.getYear() else ""
    set_date_str(date_str=date_str, target=target)


def set_date_str(date_str: str, target: dict):
    target['dc:date'] = date_str


def get_as_int(entry_key: str, entry: dict) -> int:
    v: str = (entry[entry_key]
              if entry_key in entry else None)
    return int(v) if v else None


def set_bit_depth(bit_depth: int, target: dict):
    # we avoid to set bitDepth to 0
    if bit_depth is not None and bit_depth != 0:
        target["res:bitsPerSample"] = str(bit_depth)


def set_channel_count(channel_count: int, target: dict):
    if channel_count:
        target["res:channels"] = str(channel_count)


def set_sample_rate(sample_rate: int, target: dict):
    if sample_rate:
        target["res:samplefreq"] = str(sample_rate)


def set_bit_rate(bit_rate: int, target: dict):
    if bit_rate is not None and bit_rate != 0:
        target["res:bitrate"] = str(bit_rate)


def set_size(size: int, target: dict):
    if size:
        target["res:size"] = size


def set_mimetype(mimetype: str, target: dict):
    if mimetype:
        target["res:mime"] = mimetype


def build_didlfrag(key: str, role: str, value: str):
    return f"<{key} role=\"{role}\">{html.escape(value)}</{key}>"


def set_didlfrag(didlfrag: str, target: dict):
    if didlfrag:
        target["didlfrag"] = didlfrag


def set_raw_metadata(raw_metadata_name: str, metadata_value: str, target: dict):
    if not metadata_value:
        return
    if isinstance(metadata_value, int):
        metadata_value = str(metadata_value)
    if len(metadata_value) > 0:
        target[raw_metadata_name] = metadata_value


def set_upnp_meta(metadata_name: constants.UpnpMeta, metadata_value: str, target: dict):
    set_raw_metadata(
        raw_metadata_name=f"upnp:{metadata_name.value}",
        metadata_value=metadata_value,
        target=target)


def set_upmpd_meta(metadata_name: upmpdmeta.UpMpdMeta, metadata_value: str, target: dict):
    set_raw_metadata(
        raw_metadata_name=f"upmpd:{metadata_name.value}",
        metadata_value=metadata_value,
        target=target)



# Copyright (C) 2025,2026 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
from enum import Enum
from typing import Protocol
import sqlhelper


class KeyValueCachingException(Exception):
    pass


class Version(Enum):
    V1 = "v1"


class KeyValueTableName(Enum):
    TABLE_NAME_V1 = (Version.V1, "kv_cache_v1")

    @property
    def version(self) -> Version:
        return self.value[0]

    @property
    def table_name(self) -> str:
        return self.value[1]


class KeyValueCacheColumnName(Enum):
    CREATED_TIMESTAMP = "created_timestamp"
    UPDATED_TIMESTAMP = "updated_timestamp"
    ITEM_VALUE = "value"
    ITEM_KEY = "key"
    ITEM_PARTITION = "partition"


def _get_table_by_version(version: Version) -> KeyValueTableName:
    current: KeyValueTableName
    for current in KeyValueTableName:
        if current.version == version:
            return current
    return KeyValueCachingException(f"No result for version [{version}]")


def build_create_v1_sql() -> str:
    return f"""
        CREATE TABLE {KeyValueTableName.TABLE_NAME_V1.table_name}(
        {KeyValueCacheColumnName.ITEM_PARTITION.value} VARCHAR(255),
        {KeyValueCacheColumnName.ITEM_KEY.value} VARCHAR(255),
        {KeyValueCacheColumnName.ITEM_VALUE.value}  VARCHAR(255),
        {KeyValueCacheColumnName.CREATED_TIMESTAMP.value} TIMESTAMP,
        {KeyValueCacheColumnName.UPDATED_TIMESTAMP.value} TIMESTAMP,
        PRIMARY KEY (
        {KeyValueCacheColumnName.ITEM_PARTITION.value},
        {KeyValueCacheColumnName.ITEM_KEY.value}))
"""


class KeyValueItem:

    def __init__(
            self,
            partition: str,
            key: str,
            value: str,
            created_timestamp: datetime.datetime = None,
            updated_timestamp: datetime.datetime = None):
        self.__partition: str = partition
        self.__key: str = key
        self.__value: str = value
        self.__created_timestamp: datetime.datetime = (created_timestamp
                                                       if created_timestamp
                                                       else datetime.datetime.now())
        self.__updated_timestamp: datetime.datetime = (updated_timestamp
                                                       if updated_timestamp
                                                       else self.created_timestamp)

    @property
    def partition(self) -> str:
        return self.__partition

    @property
    def key(self) -> str:
        return self.__key

    @property
    def value(self) -> str:
        return self.__value

    @property
    def created_timestamp(self) -> datetime.datetime:
        return self.__created_timestamp

    @property
    def updated_timestamp(self) -> datetime.datetime:
        return self.__updated_timestamp

    def update(self, value: str):
        self.__value = value
        self.__updated_timestamp = datetime.datetime.now()


def insert_kv_item_v1(
        sql_executor: sqlhelper.SqlExecutor,
        key_value_item: KeyValueItem,
        creation_timestamp: datetime.datetime,
        do_commit: bool = True) -> None:
    insert_values = (
        key_value_item.partition,
        key_value_item.key,
        key_value_item.value,
        creation_timestamp,
        creation_timestamp)
    insert_sql: str = sqlhelper.create_simple_insert_sql(
        table_name=KeyValueTableName.TABLE_NAME_V1.table_name,
        column_list=[
            KeyValueCacheColumnName.ITEM_PARTITION.value,
            KeyValueCacheColumnName.ITEM_KEY.value,
            KeyValueCacheColumnName.ITEM_VALUE.value,
            KeyValueCacheColumnName.CREATED_TIMESTAMP.value,
            KeyValueCacheColumnName.UPDATED_TIMESTAMP.value])
    sql_executor(
        sql=insert_sql,
        data=insert_values,
        do_commit=do_commit)


def update_kv_item_v1(
        sql_executor: sqlhelper.SqlExecutor,
        partition: str,
        key: str,
        value: str,
        update_timestamp: datetime.datetime) -> None:
    update_values = (
        value,
        update_timestamp,
        partition,
        key)
    update_sql: str = sqlhelper.create_simple_update_sql(
        table_name=KeyValueTableName.TABLE_NAME_V1.table_name,
        set_column_list=[
            KeyValueCacheColumnName.ITEM_VALUE.value,
            KeyValueCacheColumnName.UPDATED_TIMESTAMP.value],
        where_column_list=[
            KeyValueCacheColumnName.ITEM_PARTITION.value,
            KeyValueCacheColumnName.ITEM_KEY.value])
    sql_executor(
        sql=update_sql,
        data=update_values,
        do_commit=True)


def load_kv_item_v1(
        sql_selector: sqlhelper.SqlSelector,
        partition: str,
        key: str) -> KeyValueItem:
    t = (partition, key)
    q: str = sqlhelper.create_simple_select_sql(
        table_name=KeyValueTableName.TABLE_NAME_V1.table_name,
        select_column_list=[
            KeyValueCacheColumnName.CREATED_TIMESTAMP.value,
            KeyValueCacheColumnName.UPDATED_TIMESTAMP.value,
            KeyValueCacheColumnName.ITEM_VALUE.value],
        where_column_list=[
            KeyValueCacheColumnName.ITEM_PARTITION.value,
            KeyValueCacheColumnName.ITEM_KEY.value])
    rows = sql_selector(sql=q, parameters=t)
    if not rows:
        return None
    if len(rows) > 1:
        raise KeyValueCachingException(f"Multiple {KeyValueTableName.TABLE_NAME_V1.table_name} records for [{partition}] [{key}]")
    row = rows[0]
    result: KeyValueItem = KeyValueItem(
        created_timestamp=row[0],
        updated_timestamp=row[1],
        partition=partition,
        key=key,
        value=row[2])
    return result


def delete_kv_item_v1(
        sql_executor: sqlhelper.SqlExecutor,
        partition: str,
        key: str):
    t = (partition, key)
    q: str = f"""
            DELETE FROM
                {KeyValueTableName.TABLE_NAME_V1.table_name}
            WHERE
                {KeyValueCacheColumnName.ITEM_PARTITION.value} = ? AND
                {KeyValueCacheColumnName.ITEM_KEY.value} = ?
            """
    sql_executor(sql=q, data=t, do_commit=True)


class KvItemLoader(Protocol):
    def __call__(
            self,
            partition: str,
            key: str) -> KeyValueItem:
        ...


class KvItemUpdater(Protocol):
    def __call__(
            self,
            partition: str,
            key: str,
            value: str,
            update_timestamp: datetime.datetime,
            do_commit: bool) -> None:
        ...


class KvItemCreator(Protocol):
    def __call__(
            self,
            key_value_item: KeyValueItem,
            creation_timestamp: datetime.datetime,
            do_commit: bool) -> None:
        ...


def get_key_value_item(
        partition: str,
        key: str,
        kv_loader: KvItemLoader) -> KeyValueItem:
    kv_item: KeyValueItem = kv_loader(partition=partition, key=key)
    return kv_item


def put_key_value_item(
        key_value_item: KeyValueItem,
        creator: KvItemCreator,
        updater: KvItemUpdater,
        loader: KvItemLoader,
        do_commit: bool = True):
    existing: KeyValueItem = get_key_value_item(
        partition=key_value_item.partition,
        key=key_value_item.key,
        kv_loader=loader)
    if existing:
        # update persistent data
        now: datetime.datetime = datetime.datetime.now()
        updater(
            partition=key_value_item.partition,
            key=key_value_item.key,
            value=key_value_item.value,
            update_timestamp=now,
            do_commit=do_commit)
        # update cache
        existing.update(value=key_value_item.value)
    else:
        # insert
        creator(
            key_value_item=key_value_item,
            creation_timestamp=datetime.datetime.now(),
            do_commit=do_commit)



#!/usr/bin/env python3
#
# Copyright (C) 2017 J.F.Dockes
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function
import sys
import os
import json
import re
import cmdtalkplugin

import uprclsearch
import uprclindex

from upmplgutils import uplog, setidprefix
from uprclutils import waitentry
import uprclinit

setidprefix("uprcl")

#####
# Initialize communication with our parent process: pipe and method
# call dispatch

# Some of the modules we use write garbage to stdout, which messes the communication with our
# parent. Why can't people understand that this is verboten ? So we dup stdout and close it, then
# pass the right file to cmdtalk.  (hoping that none of the imports above print anything, else we'll
# have to move this code up)
_outfile = os.fdopen(os.dup(1), "w")
os.close(1)
fd = os.open("/dev/null", os.O_WRONLY)
# print("UPRCL-APP: got fd %d for /dev/null" % fd, file=sys.stderr)


# The normal system exit gets stuck on waiting for the bottle
# thread. We have nothing to really cleanup, so set up forced exit
# handler
def doexit(val):
    os._exit(val)


# Func name to method mapper
dispatcher = cmdtalkplugin.Dispatch()
# Pipe message handler
msgproc = cmdtalkplugin.Processor(dispatcher, outfile=_outfile, exitfunc=doexit)


@dispatcher.record("trackuri")
def trackuri(a):
    # This is used for plugins which generate temporary local urls
    # pointing to the microhttpd instance. The microhttpd
    # answer_to_connection() routine in plgwithslave calls 'trackuri'
    # to get a real URI to redirect to. We generate URIs which
    # directly point to our python http server, so this method should
    # never be called.
    msgproc.log("trackuri: [%s]" % a)
    raise Exception("trackuri: should not be called for uprcl!")


# objid prefix to module map
rootmap = {}


def _rootentries():
    # Build up root directory. This is our top internal structure. We
    # let the different modules return their stuff, and we take note
    # of the objid prefixes for later dispatching
    entries = []
    for treename in uprclinit.getTreesOrder():
        nents = uprclinit.getTree(treename).rootentries(uprclinit.getObjPrefix())
        for e in nents:
            rootmap[e["id"]] = treename
        entries += nents
    uplog("Browse root: rootmap now %s" % rootmap)
    return entries


def _browsedispatch(objid, bflg, offset, count):
    for id, treename in rootmap.items():
        # uplog("Testing %s against %s" % (objid, id))
        if objid.startswith(id):
            return uprclinit.getTree(treename).browse(objid, bflg, offset, count)
    raise Exception("Browse: dispatch: bad objid not in rootmap: [%s]" % objid)


@dispatcher.record("browse")
def browse(a):
    msgproc.log(f"browse: httphp [{uprclinit.getHttphp()}]   ARGS {a}")
    if "objid" not in a:
        raise Exception("No objid in args")

    objid = a["objid"]
    bflg = a["flag"] if "flag" in a else "children"

    offset = 0
    if "offset" in a:
        offset = int(a["offset"])
    count = 0
    if "count" in a:
        count = int(a["count"])

    if not objid.startswith(uprclinit.getObjPrefix()):
        raise Exception("bad objid <%s>" % objid)

    idpath = objid.replace(uprclinit.getObjPrefix(), "", 1)

    entries = []
    nocache = "1"
    try:
        if not uprclinit.initdone():
            # initdone() acquires the readlock
            entries = [
                waitentry(objid + "notready", objid, uprclinit.getHttphp()),
            ]
        else:
            initstatus, initmessage = uprclinit.initstatus()
            if not initstatus:
                entries = [
                    waitentry(
                        objid + "notready",
                        objid,
                        uprclinit.getHttphp(),
                        "Uprcl init error: " + initmessage,
                    ),
                ]
            else:
                if not idpath:
                    entries = _rootentries()
                else:
                    if len(rootmap) == 0:
                        _rootentries()
                    entries = _browsedispatch(objid, bflg, offset, count)
    finally:
        uprclinit.g_dblock.release_read()

    total = -1
    resoffs = 0
    if type(entries) == type(()):
        resoffs = entries[0]
        total = entries[1]
        entries = entries[2]
    # msgproc.log("%s" % entries)
    encoded = json.dumps(entries)
    return {"entries": encoded, "nocache": nocache, "offset": str(resoffs), "total": str(total)}


@dispatcher.record("search")
def search(a):
    msgproc.log("search: [%s]" % a)
    objid = a["objid"]
    if re.match(r"0\$uprcl\$", objid) is None:
        raise Exception("bad objid [%s]" % objid)

    upnps = a["origsearch"]
    nocache = "1"

    try:
        if not uprclinit.initdone():
            entries = [
                waitentry(objid + "notready", objid, uprclinit.getHttphp()),
            ]
        else:
            initstatus, initmessage = uprclinit.initstatus()
            if not initstatus:
                entries = [
                    waitentry(
                        objid + "notready",
                        objid,
                        uprclinit.getHttphp(),
                        "Uprcl init error: " + initmessage,
                    ),
                ]
            else:
                entries = uprclsearch.search(
                    uprclinit.getTree("folders"),
                    uprclinit.getRclConfdir(),
                    objid,
                    upnps,
                    uprclinit.getObjPrefix(),
                    uprclinit.getHttphp(),
                    uprclinit.getPathPrefix(),
                )
    finally:
        uprclinit.g_dblock.release_read()

    encoded = json.dumps(entries)
    return {"entries": encoded, "nocache": nocache}


uprclinit.uprcl_init()
msgproc.log("Uprcl running")
msgproc.mainloop()



#!/usr/bin/env python3
#
# Main program for fetching a stream URL

import sys
import logging

from StreamDecoder import StreamDecoder

class myCfg:
    def __init__(self):
        self.cf = dict()
        self.cf["url_timeout"] = "10"
    def getConfigValue(self, s):
        if s in self.cf:
            return self.cf[s]
        else:
            return None
    def setConfigValue(self, s, v):
        self.cf[s] = v


logger = logging.getLogger('upmpdcli')
logger.setLevel(logging.ERROR)
handler = logging.StreamHandler()
#handler = logging.NullHandler()
formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
handler.setFormatter(formatter)
logger.addHandler(handler)


decoder = StreamDecoder(myCfg())

urlInfo = decoder.getMediaStreamInfo(sys.argv[1])

while urlInfo is not None and urlInfo.isPlaylist():
    playlist = decoder.getPlaylist(urlInfo)
    if playlist is None or len(playlist) == 0:
        logger.error("Received empty stream from station")
        sys.exit(1)
    stream = playlist.pop(0)
    logger.info('Stream %s' % stream)
    urlInfo = decoder.getMediaStreamInfo(stream)

if urlInfo is not None:
    logger.info("Result: isplaylist %d content-type %s url %s",
                urlInfo.isPlaylist(), urlInfo.getContentType(),
                urlInfo.getUrl())
    print("%s" % urlInfo.getUrl())
else:
    logger.error(f"Ended with null urlinfo [{sys.argv[1]}]")
    print()



# Copyright (C) 2023,2024,2025 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from option_key import OptionKey


def get_option(options: dict[str, any], option_key: OptionKey) -> any:
    return (options[option_key.get_name()]
            if options and option_key.get_name() in options
            else option_key.get_default_value())


def set_option(options: dict[str, any], option_key: OptionKey, option_value: any) -> None:
    options[option_key.get_name()] = option_value



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014-2015 Thomas Amland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import print_function

import re
import sys

try:
    from urlparse import urlsplit, parse_qs
    from urllib import urlencode
except ImportError:
    from urllib.parse import urlsplit, parse_qs, urlencode


try:
    import xbmc
    import xbmcaddon

    _addon_id = xbmcaddon.Addon().getAddonInfo("id")

    def log(msg):
        msg = "[%s][routing] %s" % (_addon_id, msg)
        xbmc.log(msg, level=xbmc.LOGDEBUG)

except ImportError:

    def log(msg):
        print(msg, file=sys.stderr)


class RoutingError(Exception):
    pass


class Plugin(object):
    """
    :ivar args: The parsed query string.
    :type args: dict of byte strings
    """

    def __init__(self, base_url=None):
        self._rules = {}  # function to list of rules
        self.args = {}
        self.base_url = base_url
        if self.base_url is None:
            self.base_url = "plugin://" + xbmcaddon.Addon().getAddonInfo("id")

    def route_for(self, path):
        """
        Returns the view function for path.

        :type path: byte string.
        """
        if path.startswith(self.base_url):
            path = path.split(self.base_url, 1)[1]

        for view_fun, rules in self._rules.items():
            for rule in rules:
                if rule.match(path) is not None:
                    return view_fun
        return None

    def url_for(self, func, *args, **kwargs):
        """
        Construct and returns an URL for view function with give arguments.
        """
        if func in self._rules:
            for rule in self._rules[func]:
                path = rule.make_path(*args, **kwargs)
                if path is not None:
                    return self.url_for_path(path)
        raise RoutingError(
            "No known paths to '{0}' with args {1} and "
            "kwargs {2}".format(func.__name__, args, kwargs)
        )

    def url_for_path(self, path):
        """
        Returns the complete URL for a path.
        """
        path = path if path.startswith("/") else "/" + path
        return self.base_url + path

    def route(self, pattern):
        """Register a route."""

        def decorator(func):
            self.add_route(func, pattern)
            return func

        return decorator

    def add_route(self, func, pattern):
        """Register a route."""
        rule = UrlRule(pattern)
        if func not in self._rules:
            self._rules[func] = []
        self._rules[func].append(rule)

    def run(self, argv=sys.argv):
        if len(argv) > 2:
            self.args = parse_qs(argv[2].lstrip(b"?"))
        path = urlsplit(argv[0]).path or "/"
        self._dispatch(path)

    def redirect(self, path):
        self._dispatch(path)

    def _dispatch(self, path):
        for view_func, rules in self._rules.items():
            for rule in rules:
                kwargs = rule.match(path)
                if kwargs is not None:
                    log("Dispatching to '%s', args: %s" % (view_func.__name__, kwargs))
                    view_func(**kwargs)
                    return
        raise RoutingError('No route to path "%s"' % path)


class UrlRule(object):

    def __init__(self, pattern):
        kw_pattern = r"<(?:[^:]+:)?([A-z]+)>"
        self._pattern = re.sub(kw_pattern, "{\\1}", pattern)
        self._keywords = re.findall(kw_pattern, pattern)

        p = re.sub("<([A-z]+)>", "<string:\\1>", pattern)
        p = re.sub("<string:([A-z]+)>", "(?P<\\1>[^/]+?)", p)
        p = re.sub("<path:([A-z]+)>", "(?P<\\1>.*)", p)
        self._compiled_pattern = p
        self._regex = re.compile("^" + p + "$")

    def match(self, path):
        """
        Check if path matches this rule. Returns a dictionary of the extracted
        arguments if match, otherwise None.
        """
        # match = self._regex.search(urlsplit(path).path)
        match = self._regex.search(path)
        return match.groupdict() if match else None

    def make_path(self, *args, **kwargs):
        """Construct a path from arguments."""
        if args and kwargs:
            return None  # can't use both args and kwargs
        if args:
            # Replace the named groups %s and format
            try:
                return re.sub(r"{[A-z]+}", r"%s", self._pattern) % args
            except TypeError:
                return None

        # We need to find the keys from kwargs that occur in our pattern.
        # Unknown keys are pushed to the query string.
        url_kwargs = dict(((k, v) for k, v in kwargs.items() if k in self._keywords))
        qs_kwargs = dict(((k, v) for k, v in kwargs.items() if k not in self._keywords))

        query = "?" + urlencode(qs_kwargs) if qs_kwargs else ""
        try:
            return self._pattern.format(**url_kwargs) + query
        except KeyError:
            return None

    def __str__(self):
        return b"Rule(pattern=%s, keywords=%s)" % (self._pattern, self._keywords)



# Copyright (C) 2023,2024,2025,2026 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import config

from subsonic_connector.response import Response
from subsonic_connector.list_type import ListType
from subsonic_connector.album_list import AlbumList
from subsonic_connector.artist import Artist
from subsonic_connector.album import Album
from subsonic_connector.song import Song
from subsonic_connector.internet_radio_stations import InternetRadioStations
from subsonic_connector.search_result import SearchResult
import metadata_converter
import artist_from_album as artist_from_album
from song_data_structures import SongArtistType
from table_name import TableName
from album_metadata import AlbumMetadata
from metadata_model import AlbumMetadataModel
from album_property_key import AlbumPropertyKey
from typing import Callable
from typing import Any
import subsonic_util
import album_util
import connector_provider
from msgproc_provider import msgproc
import constants
import upmplgutils
import persistence
import sqlite3
import shutil
import time
import datetime
import os
import glob
import pathlib


def get_image_cache_path_for_pruning(www_image_path: list[str]) -> str:
    # check cache dir
    cache_dir: str = upmplgutils.getUpnpWebDocRoot(constants.PluginConstant.PLUGIN_NAME.value)
    if not cache_dir:
        msgproc.log("Cache directory is not set, cannot allow pruning.")
        return None
    candidate: pathlib.Path = pathlib.Path(cache_dir)
    # does the path actually exist?
    if not candidate.exists() or not candidate.is_dir():
        msgproc.log(f"Invalid cache path [{candidate}], cannot allow pruning.")
        return None
    # is the provided argument a valid non-empty list?
    if not www_image_path or not isinstance(www_image_path, list) or len(www_image_path) == 0:
        msgproc.log("www_image_path is not a valid list, cannot allow pruning.")
        return None
    return subsonic_util.ensure_directory(
        upmplgutils.getUpnpWebDocRoot(constants.PluginConstant.PLUGIN_NAME.value),
        www_image_path)


def prune_cache(images_static_dir: str):
    now: float = time.time()
    # list directories
    file_count: int = 0
    deleted_count: int = 0
    max_age_seconds: int = config.get_config_param_as_int(constants.ConfigParam.CACHED_IMAGE_MAX_AGE_DAYS) * (24 * 60 * 60)
    for filename in glob.glob(f"{images_static_dir}/**/*", recursive=True):
        filename_path = os.path.normpath(filename)
        file_count += 1
        time_diff_sec: float = now - os.path.getmtime(filename_path)
        if time_diff_sec >= float(max_age_seconds):
            os.remove(filename_path)
            deleted_count += 1
    msgproc.log(f"Deleted [{deleted_count}] cached images out of [{file_count}]")


def subsonic_init():
    msgproc.log(f"Subsonic [{constants.PluginConstant.PLUGIN_RELEASE.value}] Initializing ...")
    init_success: bool = False
    try:
        cache_dir: str = upmplgutils.getcachedir(constants.PluginConstant.PLUGIN_NAME.value)
        msgproc.log(f"Cache dir for [{constants.PluginConstant.PLUGIN_NAME.value}] is "
                    f"[{cache_dir}]")
        msgproc.log(f"DB version for [{constants.PluginConstant.PLUGIN_NAME.value}] is "
                    f"[{persistence.get_db_version()}]")
        persistence.purge_spurious_caches()
        known_key_list: list[str] = [x.property_key for x in AlbumPropertyKey]
        del_unknown: int = persistence.purge_unknown_album_properties(valid_property_key_list=known_key_list)
        msgproc.log(f"subsonic_init purge_unknown_album_properties deleted [{del_unknown}] records")
        purge_id_cache()
        initial_caching()
        check_supports()
        detect_anomalies()
        if config.get_config_param_as_bool(constants.ConfigParam.EXECUTE_VACUUM):
            persistence.do_vacuum()
        if config.getWebServerDocumentRoot():
            msgproc.log("WebServer is enabled ...")
            path_images_static: list[str] = config.get_webserver_path_images_static()
            images_static_dir: str = subsonic_util.ensure_directory(
                config.getWebServerDocumentRoot(),
                path_images_static)
            msgproc.log(f"Directory for static images [{images_static_dir}] created.")
            path_images_subsonic: list[str] = config.get_webserver_path_images_cache()
            images_cached_dir: str = subsonic_util.ensure_directory(
                config.getWebServerDocumentRoot(),
                path_images_subsonic)
            msgproc.log(f"Directory for cached images [{images_cached_dir}] created.")
            pkg_datadir: str = upmplgutils.getOptionValue('pkgdatadir')
            msgproc.log(f"pkg_datadir: [{pkg_datadir}]")
            src_path: str = (f"{pkg_datadir}/cdplugins/{constants.PluginConstant.PLUGIN_NAME.value}")
            src_static_images_path: str = f"{src_path}/images/static"
            for img in ["unknown-artist.svg", "unknown-cover.svg"]:
                shutil.copyfile(f"{src_static_images_path}/{img}",
                                f"{images_static_dir}/{img}")
        else:
            msgproc.log("WebServer not available")
        init_success = True
    except Exception as e:
        msgproc.log(f"Subsonic [{constants.PluginConstant.PLUGIN_RELEASE.value}] "
                    f"Initialization failed [{e}]")
    if config.get_config_param_as_bool(constants.ConfigParam.ENABLE_CACHED_IMAGE_AGE_LIMIT):
        prune_path: str = get_image_cache_path_for_pruning(config.get_webserver_path_images_cache())
        if prune_path:
            msgproc.log(f"Pruning image cache at path [{prune_path}] ...")
            prune_cache(images_static_dir=prune_path)
            msgproc.log(f"Pruned image cache at path [{prune_path}].")
    else:
        msgproc.log("Image pruning disabled.")
    msgproc.log(f"Subsonic [{constants.PluginConstant.PLUGIN_RELEASE.value}] "
                f"Initialization success: [{init_success}]")


def detect_anomalies():
    # detect_multiple_artists()
    pass


def check_supports():
    check_supports_highest()
    check_supports_internet_radios()


def check_supports_highest():
    # see if there is support for highest in getAlbumLists2
    supported: bool = False
    try:
        res: Response[AlbumList] = connector_provider.get().getAlbumList(ltype=ListType.HIGHEST, size=1)
        if res and res.isOk():
            # supported!
            supported = True
    except Exception as ex:
        msgproc.log(f"check_supports_highest highest not supported [{type(ex)}] [{ex}]")
    msgproc.log(f"highest type in getAlbumList supported: [{'yes' if supported else 'no'}]")
    if not supported:
        config.album_list_by_highest_supported = False


def check_supports_internet_radios():
    # see if there is support for highest in getAlbumLists2
    supported: bool = False
    try:
        res: Response[InternetRadioStations] = connector_provider.get().getInternetRadioStations()
        if res and res.isOk():
            # supported!
            supported = True
    except Exception as ex:
        msgproc.log(f"check_supports_highest highest not supported [{type(ex)}] [{ex}]")
    msgproc.log(f"Internet Radio stations supported: [{'yes' if supported else 'no'}]")
    if not supported:
        config.internet_radio_stations_supported = False


def wrap_action_with_connection(action: Callable[[sqlite3.Connection], Any], context: str):
    connection: sqlite3.Connection = persistence.get_working_connection(provided=None)
    res: Any = None
    try:
        res = action(connection)
    except Exception as ex:
        msgproc.log(f"{context} failed [{type(ex)}] [{ex}]")
        # don't commit
        connection.rollback()
        connection.close()
    persistence.commit(connection)
    connection.close()
    return res


class PreloadedAlbum:

    def __init__(
            self,
            album_id: str,
            song_count: int):
        self.__album_id: str = album_id
        self.__song_count: int = song_count

    @property
    def album_id(self) -> str:
        return self.__album_id

    @property
    def song_count(self) -> int:
        return self.__song_count


class PreloadAlbumsResult:

    def __init__(self):
        self.__albums: dict[str, PreloadedAlbum] = {}

    def add_album(self, album: PreloadedAlbum):
        self.__albums[album.album_id] = album

    def get_album(self, album_id: str) -> PreloadedAlbum | None:
        return self.__albums[album_id] if album_id in self.__albums else None

    @property
    def album_count(self) -> int:
        return len(self.__albums)


def get_loaded_song_list(loaded_by_album_id: dict[str, list[Song]], album_id: str) -> list[Song]:
    existing: list[Song] = loaded_by_album_id[album_id] if album_id in loaded_by_album_id else None
    if existing is None:
        existing = []
        loaded_by_album_id[album_id] = existing
    return existing


def preload_songs(connection: sqlite3.Connection, preload_albums_result: PreloadAlbumsResult):
    msgproc.log("preload_songs starting ...")
    verbose_logging: bool = config.get_config_param_as_bool(constants.ConfigParam.VERBOSE_LOGGING)
    preload_verbose_logging: bool = (verbose_logging and
                                     config.get_config_param_as_bool(constants.ConfigParam.PRELOAD_VERBOSE_LOGGING))
    start: float = time.time()
    preload_start: datetime.datetime = datetime.datetime.now()
    song_offset: int = 0
    total_stored: int = 0
    req_count: int = constants.Defaults.SUBSONIC_API_MAX_RETURN_SIZE.value
    insert_count: int = 0
    update_count: int = 0
    skip_count: int = 0
    album_id_set: set[str] = set()
    missing_album_id_set: set[str] = set()
    save_mode: persistence.SaveMode
    loaded_by_album_id: dict[str, list[Song]] = {}
    song_count_by_album_id: dict[str, int] = {}
    while True:
        res: SearchResult = connector_provider.get().search(
            query="",
            songCount=req_count,
            songOffset=song_offset,
            artistCount=0,
            albumCount=0)
        retrieved: int = len(res.getSongs())
        song: Song
        partial_insert_count: int = 0
        partial_update_count: int = 0
        partial_skip_count: int = 0
        cnt: int = 0
        for song in res.getSongs():
            album_exists: bool = song.getAlbumId() in album_id_set
            if not album_exists:
                if not song.getAlbumId() in missing_album_id_set:
                    # try loading.
                    album_md: AlbumMetadata = persistence.get_album_metadata(
                        album_id=song.getAlbumId(),
                        connection=connection)
                    if album_md:
                        album_id_set.add(song.getAlbumId())
                        album_exists = True
                    else:
                        missing_album_id_set.add(song.getAlbumId())
            cnt += 1
            if not album_exists:
                msgproc.log(f"Skipping song [{song.getId()}] (missing album [{song.getAlbumId()}])")
                partial_skip_count += 1
                continue
            if preload_verbose_logging:
                msgproc.log(f"preload_songs for [{song.getId()}] "
                            f"[{cnt}] of [{retrieved}] ([{cnt + song_offset}])...")
            try:
                _, save_mode = persistence.save_song_metadata(
                    song_metadata=metadata_converter.build_song_metadata(song=song),
                    context="preload",
                    connection=connection,
                    do_commit=False)
            except Exception as ex:
                msgproc.log(f"preload_songs error while saving song_id [{song.getId()}] [{type(ex)}] [{ex}]")
            if preload_verbose_logging:
                msgproc.log(f"preload_songs saved song_id [{song.getId()}] as [{save_mode}]")
            if save_mode == persistence.SaveMode.INSERTED:
                partial_insert_count += 1
            elif save_mode == persistence.SaveMode.UPDATED:
                partial_update_count += 1
            else:
                raise Exception(f"Invalid mode [{save_mode}]")
            persistence.save_song_album_artist_list(
                song_id=song.getId(),
                album_id=song.getAlbumId(),
                song_album_artist_list=subsonic_util.get_song_artists_by_type(
                    song=song,
                    song_artist_type=SongArtistType.SONG_ALBUM_ARTIST),
                connection=connection,
                do_commit=False)
            persistence.save_song_artist_list(
                song_id=song.getId(),
                album_id=song.getAlbumId(),
                song_artist_list=subsonic_util.get_song_artists_by_type(
                    song=song,
                    song_artist_type=SongArtistType.SONG_ARTIST),
                connection=connection,
                do_commit=False)
            persistence.save_song_contributor_list(
                song_id=song.getId(),
                album_id=song.getAlbumId(),
                song_contributor_list=subsonic_util.get_song_contributors(song=song),
                connection=connection,
                do_commit=False)
            loaded_list: list[Song] = get_loaded_song_list(loaded_by_album_id=loaded_by_album_id, album_id=song.getAlbumId())
            loaded_list.append(song)
            curr_album_song_count: int = (song_count_by_album_id[song.getAlbumId()] + 1
                                          if song.getAlbumId() in song_count_by_album_id else 1)
            song_count_by_album_id[song.getAlbumId()] = curr_album_song_count
            preloaded_album: PreloadedAlbum = (preload_albums_result.get_album(song.getAlbumId())
                                               if preload_albums_result else None)
            if preloaded_album and preloaded_album.song_count == curr_album_song_count:
                # song count for album complete, actions?
                album_properties: dict[str, list[Any]] = subsonic_util.build_album_properties_from_songs(song_list=loaded_list)
                persistence.save_album_properties(
                    album_id=song.getAlbumId(),
                    properties=album_properties,
                    connection=connection,
                    do_commit=False)
                # update album metadata?
                upd_dict: dict[AlbumMetadataModel, Any] = subsonic_util.convert_album_properties(album_properties=album_properties)
                album_path: str = album_util.get_album_path_list_joined(song_list=loaded_list)
                upd_dict[AlbumMetadataModel.ALBUM_PATH] = album_path
                if len(upd_dict) > 0:
                    persistence.update_album_metadata_table(
                        album_id=song.getAlbumId(),
                        values=upd_dict,
                        connection=connection,
                        do_commit=False)
                # purge from loaded_by_album_id
                del loaded_by_album_id[song.getAlbumId()]
        total_stored += retrieved
        insert_count += partial_insert_count
        update_count += partial_update_count
        skip_count += partial_skip_count
        msgproc.log(f"preload_songs stored [{retrieved}] "
                    f"i:[{partial_insert_count}] "
                    f"u:[{partial_update_count}] "
                    f"s:[{partial_skip_count}] "
                    f"(tot [{total_stored}] "
                    f"i:[{insert_count}] "
                    f"u:[{update_count}] "
                    f"s:[{skip_count}])")
        if (retrieved < req_count):
            # finished.
            break
        song_offset += retrieved
    # get count after loading entries
    count_before_prune: int = persistence.get_table_count(
        table_name=TableName.ALBUM_METADATA_V1,
        connection=connection)
    prune_count: int = persistence.prune_song_metadata(
        update_timestamp=preload_start,
        connection=connection,
        do_commit=False)
    elapsed: float = time.time() - start
    per_sec: float = float(total_stored) / elapsed
    msgproc.log(f"preload_songs loaded [{total_stored}] "
                f"count_before_prune [{count_before_prune}] "
                f"pruned [{prune_count}] "
                f"in [{elapsed:.3f}] "
                f"[{per_sec:.3f}] songs/sec")


def preload_albums(connection: sqlite3.Connection) -> PreloadAlbumsResult:
    msgproc.log("preload_albums starting ...")
    result: PreloadAlbumsResult = PreloadAlbumsResult()
    verbose_logging: bool = config.get_config_param_as_bool(constants.ConfigParam.VERBOSE_LOGGING)
    preload_verbose_logging: bool = (verbose_logging and
                                     config.get_config_param_as_bool(constants.ConfigParam.PRELOAD_VERBOSE_LOGGING))
    start: float = time.time()
    preload_start: datetime.datetime = datetime.datetime.now()
    album_offset: int = 0
    total_stored: int = 0
    req_count: int = constants.Defaults.SUBSONIC_API_MAX_RETURN_SIZE.value
    insert_count: int = 0
    update_count: int = 0
    save_mode: persistence.SaveMode
    while True:
        res: SearchResult = connector_provider.get().search(
            query="",
            albumCount=req_count,
            albumOffset=album_offset,
            artistCount=0,
            songCount=0)
        retrieved: int = len(res.getAlbums())
        album: Album
        partial_insert_count: int = 0
        partial_update_count: int = 0
        cnt: int = 0
        for album in res.getAlbums():
            cnt += 1
            if preload_verbose_logging:
                msgproc.log(f"preload_albums for [{album.getId()}] "
                            f"[{cnt}] of [{retrieved}] ([{cnt + album_offset}])...")
            has_album_artist: bool = album.getItem().getListByName(constants.ItemKey.ALBUM_ARTISTS.value)
            if has_album_artist:
                msgproc.log(f"preload_albums WARNING album [{album.getId()}] "
                            f"has [{constants.ItemKey.ALBUM_ARTISTS.value}] unexpectedly")
            try:
                _, save_mode = persistence.save_album_metadata(
                    album_metadata=metadata_converter.build_album_metadata(album=album),
                    context="preload",
                    connection=connection,
                    do_commit=False)
            except Exception as ex:
                msgproc.log(f"preload_albums error while saving album_id [{album.getId()}] [{type(ex)}] [{ex}]")
            if save_mode == persistence.SaveMode.INSERTED:
                partial_insert_count += 1
            elif save_mode == persistence.SaveMode.UPDATED:
                partial_update_count += 1
            else:
                raise Exception(f"Invalid mode [{save_mode}]")
            # save properties
            persistence.save_album_properties(
                album_id=album.getId(),
                properties=subsonic_util.build_album_properties(album=album),
                connection=connection,
                do_commit=False)
            # update album artists
            album_artist_list: list[artist_from_album.ArtistFromAlbum] = subsonic_util.get_artists_from_album(album=album)
            aar: int = persistence.update_album_artists(
                    album_id=album.getId(),
                    album_artists=album_artist_list,
                    connection=connection,
                    do_commit=False)
            # update album discs
            album_discs: list[subsonic_util.DiscTitle] = subsonic_util.get_disc_titles_from_album(album=album)
            ad: int = persistence.update_album_discs(
                album_id=album.getId(),
                album_discs=album_discs,
                connection=connection,
                do_commit=False)
            # genres
            album_genres: list[str] = subsonic_util.get_genres_from_album(album=album)
            ag: int = persistence.update_album_genres(
                    album_id=album.getId(),
                    album_genres=album_genres,
                    connection=connection,
                    do_commit=False)
            # record_labels
            album_record_labels: list[str] = subsonic_util.get_album_record_label_names(album=album)
            arl: int = persistence.update_album_record_labels(
                album_id=album.getId(),
                album_record_labels=album_record_labels,
                connection=connection,
                do_commit=False)
            # moods
            album_moods: list[str] = subsonic_util.get_album_moods(album=album)
            am: int = persistence.update_album_moods(
                album_id=album.getId(),
                album_moods=album_moods,
                connection=connection,
                do_commit=False)
            # release_types
            release_types: list[str] = subsonic_util.get_album_release_types(album=album).types
            art: int = persistence.update_album_release_types(
                album_id=album.getId(),
                album_release_types=release_types,
                connection=connection,
                do_commit=False)
            result.add_album(album=PreloadedAlbum(album_id=album.getId(), song_count=album.getSongCount()))
            if preload_verbose_logging:
                msgproc.log(f"preload_albums saved album_id [{album.getId()}] as [{save_mode}] "
                            f"artists [{aar}] "
                            f"discs [{ad}] "
                            f"genres [{ag}] "
                            f"record labels [{arl}] "
                            f"moods [{am}] "
                            f"release types [{art}]")
        total_stored += retrieved
        insert_count += partial_insert_count
        update_count += partial_update_count
        msgproc.log(f"preload_albums stored [{retrieved}] "
                    f"ins [{partial_insert_count}] "
                    f"upd [{partial_update_count}] "
                    f"(total [{total_stored}] "
                    f"ins [{insert_count}] "
                    f"upd [{update_count}])")
        if (retrieved < req_count):
            # finished.
            break
        album_offset += retrieved
    # get count after loading entries
    count_before_prune: int = persistence.get_table_count(
        table_name=TableName.ALBUM_METADATA_V1,
        connection=connection)
    # prune
    prune_count: int = persistence.prune_album_metadata(
        update_timestamp=preload_start,
        connection=connection,
        do_commit=False)
    elapsed: float = time.time() - start
    per_sec: float = float(total_stored) / elapsed
    msgproc.log(f"preload_albums loaded [{total_stored}] "
                f"count_before_prune [{count_before_prune}] "
                f"pruned [{prune_count}] "
                f"in [{elapsed:.3f}] "
                f"[{per_sec:.3f}] albums/sec")
    return result


def preload_artists(connection: sqlite3.Connection):
    msgproc.log("preload_artists starting ...")
    start: float = time.time()
    preload_start: datetime.datetime = datetime.datetime.now()
    msgproc.log("preload_artists connection created ...")
    artist_offset: int = 0
    total_stored: int = 0
    req_count: int = constants.Defaults.SUBSONIC_API_MAX_RETURN_SIZE.value
    while True:
        res: SearchResult = connector_provider.get().search(
            query="",
            artistCount=req_count,
            artistOffset=artist_offset,
            albumCount=0,
            songCount=0)
        retrieved: int = len(res.getArtists())
        artist: Artist
        for artist in res.getArtists():
            artist_roles: list[str] = subsonic_util.get_artist_roles(artist=artist)
            artist_metadata: persistence.ArtistMetadata = metadata_converter.build_artist_metadata(artist=artist)
            persistence.save_artist_metadata(
                artist_metadata=artist_metadata,
                connection=connection,
                do_commit=False)
            persistence.update_artist_roles(
                artist_id=artist.getId(),
                artist_roles=artist_roles,
                connection=connection,
                do_commit=False)
        total_stored += retrieved
        msgproc.log(f"preload_artists stored [{retrieved}] (total [{total_stored}])")
        if (retrieved < req_count):
            # finished.
            break
        artist_offset += retrieved
    # prune
    prune_count: int = persistence.prune_artist_metadata(
        update_timestamp=preload_start,
        connection=connection,
        do_commit=False)
    elapsed: float = time.time() - start
    per_sec: float = float(total_stored) / elapsed
    msgproc.log(f"preload_artists loaded [{total_stored}] "
                f"pruned [{prune_count}] "
                f"in [{elapsed:.3f}] "
                f"[{per_sec:.3f}] artists/sec")


def initial_caching():
    preload_success: bool = True
    preload_start: float = time.time()
    initial_caching_start: datetime.datetime = datetime.datetime.fromtimestamp(preload_start)
    preload_max_delta_sec: int = config.get_config_param_as_int(constants.ConfigParam.PRELOAD_MAX_DELTA_SEC)
    preload_artists_enabled: bool = config.get_config_param_as_bool(constants.ConfigParam.PRELOAD_ARTISTS)
    preload_albums_enabled: bool = preload_artists_enabled and config.get_config_param_as_bool(constants.ConfigParam.PRELOAD_ALBUMS)
    preload_songs_enabled: bool = preload_albums_enabled and config.get_config_param_as_bool(constants.ConfigParam.PRELOAD_SONGS)
    msgproc.log(f"initial_caching preload artists [{preload_artists_enabled}] "
                f"preload albums [{preload_albums_enabled}] "
                f"preload songs [{preload_songs_enabled}]")
    if preload_artists_enabled:
        oldest: float | None = persistence.get_oldest_metadata(table_name=TableName.ARTIST_METADATA_V1)
        artist_initial_count: int = persistence.get_table_count(
            table_name=TableName.ARTIST_METADATA_V1)
        msgproc.log(f"initial_caching {TableName.ARTIST_METADATA_V1.value} "
                    f"current count [{artist_initial_count}] "
                    f"oldest [{oldest}]")
        if oldest is None or (initial_caching_start - oldest).total_seconds() > preload_max_delta_sec:
            try:
                wrap_action_with_connection(
                    action=preload_artists,
                    context="preload_artists")
            except Exception as ex:
                msgproc.log(f"preload_artists failed [{type(ex)}] [{ex}]")
        else:
            msgproc.log("initial_caching skipping preload_artists because "
                        f"start - oldest [{(initial_caching_start - oldest).total_seconds()}] <= "
                        f"max_delta [{preload_max_delta_sec}]")
    preload_albums_result: PreloadAlbumsResult = None
    if preload_albums_enabled:
        oldest = persistence.get_oldest_metadata(table_name=TableName.ALBUM_METADATA_V1)
        album_initial_count: int = persistence.get_table_count(table_name=TableName.ALBUM_METADATA_V1)
        msgproc.log(f"initial_caching {TableName.ALBUM_METADATA_V1.value} "
                    f"current count [{album_initial_count}] "
                    f"oldest [{oldest}]")
        if oldest is None or (initial_caching_start - oldest).total_seconds() > preload_max_delta_sec:
            try:
                preload_albums_result = wrap_action_with_connection(
                    action=preload_albums,
                    context="preload_albums")
                preload_success = True
            except Exception as ex:
                msgproc.log(f"preload_albums failed [{type(ex)}] [{ex}]")
        else:
            msgproc.log("initial_caching skipping preload_albums because "
                        f"start - oldest [{(initial_caching_start - oldest).total_seconds()}] <= "
                        f"max_delta [{preload_max_delta_sec}]")
    if preload_songs_enabled:
        msgproc.log(f"Preloaded [{preload_albums_result.album_count if preload_albums_result else 0}] albums, "
                    "starting song preload")
        oldest = persistence.get_oldest_metadata(table_name=TableName.SONG_METADATA_V1)
        song_initial_count: int = persistence.get_table_count(table_name=TableName.SONG_METADATA_V1)
        msgproc.log(f"initial_caching {TableName.SONG_METADATA_V1.value} "
                    f"current count [{song_initial_count}] "
                    f"oldest [{oldest}]")
        if oldest is None or (initial_caching_start - oldest).total_seconds() > preload_max_delta_sec:
            try:
                wrap_action_with_connection(
                    action=lambda x: preload_songs(x, preload_albums_result),
                    context="preload_songs")
            except Exception as ex:
                msgproc.log(f"preload_songs failed [{type(ex)}] [{ex}]")
        else:
            msgproc.log("initial_caching skipping preload_songs because "
                        f"start - oldest [{(initial_caching_start - oldest).total_seconds()}] <= "
                        f"max_delta [{preload_max_delta_sec}]")
    preload_elapsed: float = time.time() - preload_start
    if preload_artists_enabled or preload_albums_enabled or preload_songs_enabled:
        msgproc.log(f"initial_caching completed with success [{preload_success}] "
                    f"in [{preload_elapsed:.3f}]")


def purge_id_cache():
    if config.get_config_param_as_bool(constants.ConfigParam.PURGE_IDENTIFIER_CACHE):
        persistence.purge_id_cache()



# Copyright (C) 2026 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import json
import codec


class ValueHolder:

    def __init__(self, value: str):
        self.__value: str = value
    
    @property
    def value(self) -> str:
        return self.__value
    

def encode_value_holder(value: str) -> str:
    value_holder: ValueHolder = ValueHolder(value=value)
    d: dict[str, str] = {"value": value_holder.value}
    return codec.base64_encode(json.dumps(d))

def decode_value_holder(encoded: str) -> str:
    decoded: str = codec.base64_decode(encoded)
    d: dict[str, str] = json.loads(decoded)
    value_holder: ValueHolder = ValueHolder(value=d["value"])
    return value_holder.value


# Copyright (C) 2023,2024,2025 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum


class KindType:

    ARTIST = "artist"
    ALBUM = "album"
    TRACK = "track"


class SearchType(Enum):
    ALBUM = 0, "album"
    ARTIST = 1, "artist"
    TRACK = 2, "track"
    TITLE = 3, "title"

    def __init__(
            self,
            num: int,
            element_name: str):
        self.num: int = num
        self.element_name: str = element_name

    def getName(self):
        return self.element_name


# duplicate check
name_checker_set: set[str] = set()
id_checker_set: set[int] = set()
for v in SearchType:
    if v.getName() in name_checker_set:
        raise Exception(f"Duplicated name [{v.getName()}]")
    if v.value[0] in id_checker_set:
        raise Exception(f"Duplicated id [{v.value[0]}]")
    name_checker_set.add(v.getName())
    id_checker_set.add(v.value[0])



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014 Thomas Amland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from __future__ import unicode_literals


def sortmodellist(l):
    l.sort(key=lambda e: e.name)
    return l


class Model(object):
    id = None
    name = None
    image = None

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


class User(Model):
    upnpclass = "object.container"
    name = "Unknown"


class Album(Model):
    upnpclass = "object.container.album.musicAlbum"
    name = "Unknown"
    artist = None
    artists = []
    num_tracks = -1
    description = None
    duration = -1
    release_date = None
    available = True


class Artist(Model):
    upnpclass = "object.container.person.musicArtist"
    name = "Unknown"
    role = None


class Playlist(Model):
    upnpclass = "object.container.playlistContainer"
    # Using the proper playlistContainer type this confuses e.g. Kazoo
    # upnpclass = "object.container.playlistContainer"
    # We create a bogus artist with the playlist owner name when available
    artist = None
    description = None
    num_tracks = -1
    duration = -1
    # For spotify: wants an userid to retrieve a playlist's tracks
    userid = None


class Track(Model):
    upnpclass = "object.item.audioItem.musicTrack"
    duration = -1
    track_num = -1
    disc_num = 1
    popularity = -1
    artist = None
    artists = []
    album = None
    available = True
    description = None
    mime = None
    samplefreq = None
    bitdepth = None
    channels = None


class Category(Model):
    upnpclass = "object.container"


class Genre(Model):
    upnpclass = "object.container"


class SearchResult(Model):
    artists = []
    albums = []
    tracks = []
    playlists = []



"""
    qobuz.api.raw
    ~~~~~~~~~~~~~

    Our base api, all method are mapped like in <endpoint>_<method>
    see Qobuz API on GitHub (https://github.com/Qobuz/api-documentation)

    :part_of: xbmc-qobuz
    :copyright: (c) 2012 by Joachim Basmaison, Cyril Leclerc
    :license: GPLv3, see LICENSE for more details.
"""

import sys
import time
import math
import hashlib
import socket
import binascii
from itertools import cycle
import requests
from . import spoofbuz

socket.timeout = 5

_loglevel = 3


def debug(s):
    if _loglevel >= 4:
        print("%s" % s, file=sys.stderr)


def warn(s):
    if _loglevel >= 3:
        print("%s" % s, file=sys.stderr)


class RawApi(object):

    def __init__(self, appid, configvalue):
        if appid and configvalue:
            self.configvalue = configvalue
            self.appid = appid
            self.__set_s4()
        else:
            self.spoofer = spoofbuz.Spoofer()
            self.appid = self.spoofer.getAppId()

        self.version = "0.2"
        self.baseUrl = "https://www.qobuz.com/api.json/"
        self.user_auth_token = None
        self.user_id = None
        self.error = None
        self.status_code = None
        self._baseUrl = self.baseUrl + self.version
        self.session = requests.Session()
        self.error = None

    def _api_error_string(self, request, url="", params={}, json=""):
        return (
            "{reason} (code={status_code})\n"
            "url={url}\nparams={params}"
            "\njson={json}".format(
                reason=request.reason,
                status_code=self.status_code,
                url=url,
                params=str(["%s: %s" % (k, v) for k, v in params.items()]),
                json=str(json),
            )
        )

    def _check_ka(self, ka, mandatory, allowed=[]):
        """Checking parameters before sending our request
        - if mandatory parameter is missing raise error
        - if a given parameter is neither in mandatory or allowed
        raise error
        """
        for label in mandatory:
            if not label in ka:
                raise Exception("Qobuz: missing parameter [%s]" % label)
        for label in ka:
            if label not in mandatory and label not in allowed:
                raise Exception("Qobuz: invalid parameter [%s]" % label)
        # Having no parameters set triggers a problem in the pyrequests/Qobuz dialog, don't know
        # where the bug is, but it results in a 411 (length required). So set a very high limit if
        # nothing is set
        noparams = True
        for label in ka:
            if ka[label]:
                noparams = False
                break
        if noparams:
            ka["limit"] = "10000"

    def __set_s4(self):
        """appid and associated secret is for this app usage only
        Any use of the API implies your full acceptance of the
        General Terms and Conditions
        (http://www.qobuz.com/apps/api/QobuzAPI-TermsofUse.pdf)
        """
        s3b = self.configvalue.encode("ASCII")
        s3s = binascii.a2b_base64(s3b)
        bappid = self.appid.encode("ASCII")
        a = cycle(bappid)
        b = zip(s3s, a)
        self.s4 = b"".join((x ^ y).to_bytes(1, byteorder="big") for (x, y) in b)
        # print("S4: %s"% self.s4.decode('ASCII'), file=sys.stderr)

    def __unset_s4(self, id, sec):
        a = cycle(id)
        b = zip(sec, a)
        bs4 = b"".join((x ^ y).to_bytes(1, byteorder="big") for (x, y) in b)
        value = binascii.b2a_base64(bs4)
        return value

    def _api_request(self, params, uri, **opt):
        """Qobuz API HTTP get request
        Arguments:
        params:    parameters dictionary
        uri   :    service/method
        opt   :    Optional named parameters
                    - noToken=True/False
                    - useGet=True/False

        Return None if something went wrong
        Return raw data from qobuz on success as dictionary

        * on error you can check error and status_code

        Example:

            ret = api._api_request({'username':'foo',
                              'password':'bar'},
                             'user/login', noToken=True)
            print('Error: %s [%s]' % (api.error, api.status_code))

        This should produce something like:
        Error: [200]
        Error: Bad Request [400]
        """
        self.error = ""
        self.status_code = None
        url = self._baseUrl + uri
        useToken = False if (opt and "noToken" in opt) else True
        useGet = True if (opt and "useGet" in opt) else False
        headers = {
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:67.0) Gecko/20100101 Firefox/67.0"
        }
        if useToken and self.user_auth_token:
            headers["X-User-Auth-Token"] = self.user_auth_token
        headers["X-App-Id"] = self.appid
        r = None
        op = "GET" if useGet else "POST"
        # warn(f"{op} {url} params {params} headers {headers}")
        debug(f"{op} {url} params {params}")
        try:
            if useGet:
                r = self.session.get(url, params=params, headers=headers)
            else:
                r = self.session.post(url, data=params, headers=headers)
        except:
            self.error = "Post request fail"
            warn(self.error)
            return None
        debug(f"status_code: {r.status_code}\nheaders: {r.headers}\ncontent: {r.content}")
        self.status_code = int(r.status_code)
        if self.status_code != 200:
            self.error = self._api_error_string(r, url, params)
            warn(self.error)
            return None
        if not r.content:
            self.error = "Request return no content"
            warn(self.error)
            return None

        """Retry get if connexion fail"""
        try:
            response_json = r.json()
        except Exception as e:
            warn("Json loads failed to load... retrying!\n{}", repr(e))
            try:
                response_json = r.json()
            except:
                self.error = "Failed to load json two times...abort"
                warn(self.error)
                return None
        status = None
        try:
            status = response_json["status"]
        except:
            pass
        if status == "error":
            self.error = self._api_error_string(r, url, params, response_json)
            warn(self.error)
            return None
        return response_json

    def logout(self):
        self.user_auth_token = None
        self.user_id = None
        self.logged_on = None

    def user_login(self, **ka):
        self._check_ka(ka, ["username", "password"], ["email"])
        data = self._api_request(ka, "/user/login", noToken=True)
        if (
            not data
            or not "user" in data
            or not "credential" in data["user"]
            or not "id" in data["user"]
            or not "parameters" in data["user"]["credential"]
        ):
            warn("/user/login returns %s" % data)
            self.logout()
            return None
        if not data["user"]["credential"]["parameters"]:
            warn("Free accounts are not eligible to download tracks.")
            return None
        self.user_id = data["user"]["id"]
        self.user_auth_token = data["user_auth_token"]
        self.label = data["user"]["credential"]["parameters"]["short_label"]
        debug("Membership: {}".format(self.label))
        data["user"]["email"] = ""
        data["user"]["firstname"] = ""
        data["user"]["lastname"] = ""
        self.setSec()
        return data

    def setSec(self):
        global _loglevel
        savedloglevel = _loglevel
        _loglevel = 1
        for value in self.spoofer.getSecrets().values():
            self.s4 = value.encode("utf-8")
            if self.userlib_getAlbums(sec=self.s4) is not None:
                # debug("SECRET [%s]"%self.s4)
                _loglevel = savedloglevel
                return
        _loglevel = savedloglevel

    def track_get(self, **ka):
        self._check_ka(ka, ["track_id"])
        return self._api_request(ka, "/track/get")

    def track_getFileUrl(self, intent="stream", **ka):
        self._check_ka(ka, ["format_id", "track_id"])
        ts = str(time.time())
        track_id = str(ka["track_id"])
        fmt_id = str(ka["format_id"])
        stringvalue = (
            "trackgetFileUrlformat_id" + fmt_id + "intent" + intent + "track_id" + track_id + ts
        )
        stringvalue = stringvalue.encode("ASCII")
        stringvalue += self.s4
        rq_sig = str(hashlib.md5(stringvalue).hexdigest())
        params = {
            "format_id": fmt_id,
            "intent": intent,
            "request_ts": ts,
            "request_sig": rq_sig,
            "track_id": track_id,
        }
        return self._api_request(params, "/track/getFileUrl")

    def userlib_getAlbums(self, **ka):
        ts = str(time.time())
        r_sig = "userLibrarygetAlbumsList" + str(ts) + str(ka["sec"])
        r_sig_hashed = hashlib.md5(r_sig.encode("utf-8")).hexdigest()
        params = {
            "app_id": self.appid,
            "user_auth_token": self.user_auth_token,
            "request_ts": ts,
            "request_sig": r_sig_hashed,
        }
        return self._api_request(params, "/userLibrary/getAlbumsList")

    # Currently unused. Check that it works ?
    def track_search(self, **ka):
        self._check_ka(ka, ["query"], ["limit"])
        return self._api_request(ka, "/track/search")

    def album_get(self, **ka):
        self._check_ka(ka, ["album_id"], ["extra", "limit", "offset"])
        # As of around sept 2024, using a POST for this does not work any more (always return the
        # same album, not the one requested. Probably an inadvertant change on the Qobuz side. The
        # WEB player uses a GET
        return self._api_request(ka, "/album/get", useGet=True)

    def album_getFeatured(self, **ka):
        self._check_ka(ka, [], ["type", "genre_ids", "limit", "offset"])
        return self._api_request(ka, "/album/getFeatured", useGet=True)

    def favorite_getUserFavorites(self, **ka):
        self._check_ka(ka, [], ["user_id", "type", "limit", "offset"])
        return self._api_request(ka, "/favorite/getUserFavorites")

    def playlist_get(self, **ka):
        self._check_ka(ka, ["playlist_id"], ["extra", "limit", "offset"])
        return self._api_request(ka, "/playlist/get", useGet=True)

    def playlist_getFeatured(self, **ka):
        # type is 'last-created' or 'editor-picks'
        self._check_ka(ka, ["type"], ["genre_ids", "tags", "limit", "offset"])
        for k in ("tags", "genre_ids"):
            if k in ka and ka[k] == "None":
                del ka[k]
        return self._api_request(ka, "/playlist/getFeatured", useGet=True)

    def playlist_getUserPlaylists(self, **ka):
        self._check_ka(ka, [], ["user_id", "username", "order", "offset", "limit"])
        if not "user_id" in ka and not "username" in ka:
            ka["user_id"] = self.user_id
        return self._api_request(ka, "/playlist/getUserPlaylists")

    def artist_getSimilarArtists(self, **ka):
        self._check_ka(ka, ["artist_id"], ["limit", "offset"])
        return self._api_request(ka, "/artist/getSimilarArtists", useGet=True)

    def artist_get(self, **ka):
        self._check_ka(ka, ["artist_id"], ["extra", "limit", "offset"])
        return self._api_request(ka, "/artist/get", useGet=True)

    def genre_list(self, **ka):
        self._check_ka(ka, [], ["parent_id", "limit", "offset"])
        return self._api_request(ka, "/genre/list")

    def label_list(self, **ka):
        self._check_ka(ka, [], ["limit", "offset"])
        return self._api_request(ka, "/label/list")

    def catalog_search(self, **ka):
        # type may be 'tracks', 'albums', 'artists' or 'playlists'
        self._check_ka(ka, ["query"], ["type", "offset", "limit"])
        return self._api_request(ka, "/catalog/search", useGet=True)

    #### 2024-10 Except for search the /catalog/ methods still work but they're not used by the site
    #### afaics, replaced by /albums/getFeatured, /playlists/getFeatured
    def catalog_getFeatured(self, **ka):
        return self._api_request(ka, "/catalog/getFeatured")

    def catalog_getFeaturedTypes(self, **ka):
        return self._api_request(ka, "/catalog/getFeaturedTypes")



# Copyright (C) 2023,2024,2025 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum
from album_util import MultiCodecAlbum


class OptionKey(Enum):
    SKIP_TRACK_NUMBER = 0, "skip-track-number", False
    APPEND_ARTIST_IN_ALBUM_TITLE = 1, "append-artist-in-album-title", True
    FORCE_TRACK_NUMBER = 2, "force-track-number", None
    MULTI_CODEC_ALBUM = 4, "multi-codec-album", MultiCodecAlbum.NO
    SKIP_ART = 5, "skip-art", False
    OFFSET = 6, "offset", 0
    # integer to prepend e.g. pass 3 -> [03] album_title instead of album_title
    PREPEND_ENTRY_NUMBER_IN_ALBUM_TITLE = 8, "prepend-entry-number-in-album-title", None
    FORCE_RELOAD_ALBUM_QUALITY_INFO = 9, "force-load-quality-badge", False
    USE_LAST_FOR_NEXT = 11, "use-last-for-next", False
    SKIP_ARTIST_ID = 12, "skip-artist-id", None
    SEARCH_RESULT = 13, "search-result", False

    def __init__(
            self,
            num: int,
            element_name: str,
            default_value: any):
        self.num: int = num
        self.element_name: str = element_name
        self.default_value: any = default_value

    def get_name(self) -> str:
        return self.element_name

    def get_default_value(self) -> any:
        return self.default_value


# duplicate check
name_checker_set: set[str] = set()
id_checker_set: set[int] = set()
for v in OptionKey:
    if v.get_name() in name_checker_set:
        raise Exception(f"Duplicated name [{v.get_name()}]")
    if v.value[0] in id_checker_set:
        raise Exception(f"Duplicated id [{v.value[0]}]")
    name_checker_set.add(v.get_name())
    id_checker_set.add(v.value[0])



# Copyright (C) 2017 J.F.Dockes
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import os
import conftree
import threading
import subprocess
import time
import traceback
from timeit import default_timer as timer

from rwlock import ReadWriteLock

from uprclfolders import Folders
from uprcluntagged import Untagged
from uprclplaylists import Playlists
from uprcltags import Tagged
import uprclsearch
import uprclindex
from uprclhttp import runbottle
import minimconfig
import uprclutils

from upmplgutils import uplog, getcachedir, getOptionValue
from conftree import stringToStrings

### Configuration stuff
_g_rclconfdir = ""
_g_pathprefix = ""
_g_httphp = ""
_g_friendlyname = "UpMpd-mediaserver"
# Prefix for object Ids. This must be consistent with what
# contentdirectory.cxx does
_g_myprefix = "0$uprcl$"
g_minimconfig = None

### Index update status
# Running state: ""/"Updating"/"Rebuilding"
g_initrunning = ""
# Completion status: ok/notok
g_initstatus = False
# Possible error message if not ok
g_initmessage = ""

### Data created during initialisation
_g_trees = {}
_g_trees_order = ["folders", "playlists", "tags", "untagged"]

g_dblock = ReadWriteLock()


def getObjPrefix():
    return _g_myprefix


def getPathPrefix():
    return _g_pathprefix


def getHttphp():
    return _g_httphp


def getRclConfdir():
    return _g_rclconfdir


def getFriendlyname():
    return _g_friendlyname


def getTree(treename):
    return _g_trees[treename]


def getTreesOrder():
    return _g_trees_order


def _reset_index():
    _update_index(True)


def initdone():
    g_dblock.acquire_read()
    if g_initrunning:
        return False
    else:
        return True


def initstatus():
    return (g_initstatus, g_initmessage)


def updaterunning():
    return g_initrunning


# Create or update Recoll index, then read and process the data. This runs in a separate thread, and
# signals startup/completion by setting/unsetting the g_initrunning flag.
#
# While this is running, or after a failure any access to the root container from a Control Point
# will display either an "Initializing" or error message.
def _update_index(rebuild=False):
    uplog("Creating/updating index in %s for %s" % (_g_rclconfdir, g_rcltopdirs))

    # We take the writer lock, making sure that no browse/search thread are active, then set the
    # busy flag and release the lock. This allows future browse operations to signal the condition
    # to the user instead of blocking (if we kept the write lock).
    global g_initrunning, _g_trees, g_initstatus, g_initmessage
    g_dblock.acquire_write()
    g_initrunning = "Rebuilding" if rebuild else "Updating"
    g_dblock.release_write()
    uplog("_update_index: initrunning set")

    try:
        start = timer()
        noindexing = conftree.valToBool(getOptionValue("uprclnoindexing"))
        if not noindexing:
            uprclindex.runindexer(_g_rclconfdir, g_rcltopdirs, rebuild=rebuild)
            # Wait for indexer
            while not uprclindex.indexerdone():
                time.sleep(0.5)
            fin = timer()
            uplog("Indexing took %.2f Seconds" % (fin - start))

        folders = Folders(_g_rclconfdir, _g_httphp, _g_pathprefix)
        uprclutils.importfolders(folders)
        untagged = Untagged(folders.rcldocs(), _g_httphp, _g_pathprefix)
        playlists = Playlists(_g_rclconfdir, folders.rcldocs(), _g_httphp, _g_pathprefix)
        tagged = Tagged(folders, _g_httphp, _g_pathprefix)
        newtrees = {}
        newtrees["folders"] = folders
        newtrees["untagged"] = untagged
        newtrees["playlists"] = playlists
        newtrees["tags"] = tagged
        _g_trees = newtrees
        g_initstatus = True
        uplog("Init done")
    except Exception as ex:
        traceback.print_exc()
        g_initstatus = False
        g_initmessage = str(ex)
        uplog(f"Initialisation failed with: {g_initmessage}")
    finally:
        g_dblock.acquire_write()
        g_initrunning = ""
        g_dblock.release_write()


# This is called from uprcl-app when starting up, before doing anything else. We read configuration
# data, then start two threads: the permanent HTTP server and the index update thread.
def uprcl_init():

    global _g_pathprefix, g_initstatus, g_initmessage

    #######
    # Acquire configuration data.

    # We get the path prefix from an environment variable set by our parent upmpdcli. It would
    # typically be something like "/uprcl". It's used for dispatching URLs to the right plugin for
    # processing. We strip it whenever we need a real file path.
    if "UPMPD_PATHPREFIX" not in os.environ:
        raise Exception("No UPMPD_PATHPREFIX in environment")
    _g_pathprefix = os.environ["UPMPD_PATHPREFIX"]
    if "UPMPD_CONFIG" not in os.environ:
        raise Exception("No UPMPD_CONFIG in environment")
    global _g_friendlyname
    if "UPMPD_FNAME" in os.environ:
        _g_friendlyname = os.environ["UPMPD_FNAME"]

    global g_minimconfig
    minimcfn = getOptionValue("uprclminimconfig")
    g_minimconfig = minimconfig.MinimConfig(minimcfn)

    global _g_httphp
    _g_httphp = getOptionValue("uprclhostport")
    if _g_httphp is None:
        if "UPMPD_HTTPHOSTPORT" not in os.environ:
            raise Exception("No UPMPD_HTTPHOSTPORT in environment")
        upmpdhttphp = os.environ["UPMPD_HTTPHOSTPORT"]
        ip = upmpdhttphp.split(":")[0]
        port = getOptionValue("uprclport")
        if port is None:
            port = "9090"
        _g_httphp = ip + ":" + port
    uplog("uprcl: serving files on %s" % _g_httphp)

    global _g_rclconfdir
    _g_rclconfdir = getOptionValue("uprclconfdir")
    _g_rclconfdir = getcachedir("uprcl", forcedpath=_g_rclconfdir)
    uplog("uprcl: cachedir: %s" % _g_rclconfdir)

    global g_rcltopdirs
    g_rcltopdirs = getOptionValue("uprclmediadirs")
    if g_rcltopdirs:
        g_rcltopdirs = conftree.stringToStrings(g_rcltopdirs)
    else:
        g_rcltopdirs = g_minimconfig.getcontentdirs()
    if not g_rcltopdirs:
        g_initstatus = False
        g_initmessage = "No media directories were set in configuration"
        return

    # g_rcltopdirs is now a list, check the elements
    goodpthlist = []
    for dir in g_rcltopdirs:
        if not os.path.isdir(dir):
            uplog(f"uprcl: [{dir}] is not accessible")
        else:
            goodpthlist.append(dir)
    if not goodpthlist:
        g_initstatus = False
        g_initmessage = "No accessible media directories in configuration"
        return

    host, port = _g_httphp.split(":")

    # Turn g_rcltopdirs back into a string, that's how it's used by runindexer
    g_rcltopdirs = conftree.stringsToString(goodpthlist)

    start_index_update()

    # Start the bottle app. It's both the control/config interface and the file streamer
    httpthread = threading.Thread(
        target=runbottle,
        kwargs={"host": host, "port": int(port),
                "topdirs": goodpthlist, "pathprefix": _g_pathprefix},
    )
    httpthread.daemon = True
    httpthread.start()

    uplog("Init started")


# This is called from the Bottle Web UI interface for requesting an index update or rebuild
def start_index_update(rebuild=False):
    try:
        # initdone() acquires the reader lock
        if not initdone():
            return
        targ = _reset_index if rebuild else _update_index
        idxthread = threading.Thread(target=targ)
        idxthread.daemon = True
    finally:
        # We need to release the reader lock before starting the index
        # update operation (which needs a writer lock), so there is a
        # small window for mischief. I would be concerned if this was
        # a highly concurrent or critical app, but here, not so
        # much...
        g_dblock.release_read()
    idxthread.start()


# Return list of tags from minim indexTags settings. These can contain custom tags,
# not part of our predefined lists, which need to be stored in the resultstore.
def allMinimTags():
    fields = []
    try:
        storedtags = g_minimconfig.getindextags()
        for f in g_minimconfig.getitemtags():
            storedtags[f] = ""
        for v, d in storedtags:
            if v.lower() == "none":
                break
            fields.append(v.lower())
    except:
        pass
    return fields
